;;;; Copyright (c) 2004 David Lichteblau <david@lichteblau.com>
;;;;
;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :jdii)


;;;; Classes

(define-class client-vm (jdi:virtual-machine jdwp-connection)
  ((event-queue :initform (make-simple-queue))
   (event-requests :initform '())
   base-directory
   class-path
   bootstrap-class-path
   cached-classes
   primitive-classes
   jdi:description
   jdi:name
   jdi:jdwp-version
   jdi:version
   capabilities))

(defmethod initialize-instance :after ((instance client-vm) &key)
  (fetch-id-sizes instance)
  (fetch-vm-description instance)
  (fetch-capabilities instance)
  (setf (primitive-classes instance)
	(with-collector ()
	  (flet ((make-primitive-type (signature name)
		   (make-instance 'jdi:primitive-type
		     :name name
		     :signature signature)))
	    (collect (make-primitive-type "B" "byte"))
	    (collect (make-primitive-type "C" "char"))
	    (collect (make-primitive-type "D" "double"))
	    (collect (make-primitive-type "F" "float"))
	    (collect (make-primitive-type "I" "int"))
	    (collect (make-primitive-type "J" "long"))
	    (collect (make-primitive-type "S" "short"))
	    (collect (make-primitive-type "V" "void"))
	    (collect (make-primitive-type "Z" "boolean")))))
  ;; We need an up-to-date class list and depend on having the classes
  ;; already available during decoding.
  (jdi:request-event instance (make-instance 'jdi:class-prepare-request))
  (jdi:all-classes instance))

(define-class mirror ()
  ((vm :reader jdi:virtual-machine)
   id))

(define-class reference-type-mirror (mirror jdi:reference-type)
  (signature methods fields cached-superclass cached-interfaces))

(defmethod print-object ((object reference-type-mirror) stream)
  (print-unreadable-object (object stream :type t :identity nil)
    (format stream "~A #~D" (jdi:signature object) (id object))))

(define-class array-type-mirror (reference-type-mirror jdi:array-type) ())
(define-class class-type-mirror (reference-type-mirror jdi:class-type) ())
(define-class interface-type-mirror (reference-type-mirror jdi:interface-type)
  ())

(define-class member-mirror (mirror jdi:member)
  (jdi:name jdi:signature jdi:declaring-type jdi:modifiers))

(define-class field-mirror (member-mirror jdi:field) ())

(defmethod print-object ((object field-mirror) stream)
  (print-unreadable-object (object stream :type t :identity nil)
    (format stream "~A ~A.~A #~D"
	    (jni-to-java (jdi:signature object))
	    (jdi:name (jdi:declaring-type object))
	    (jdi:name object)
	    (id object))))

(define-class method-mirror (member-mirror jdi:method) ())

(defmethod print-object ((object method-mirror) stream)
  (print-unreadable-object (object stream :type t :identity nil)
    (format stream "~A.~A~A #~D"
	    (jdi:name (jdi:declaring-type object))
	    (jdi:name object)
	    (jdi:signature object)
	    (id object))))

(define-class object-reference (jdi:object mirror) ())
(define-class array-reference (jdi:array object-reference) ())
(define-class thread-mirror (jdi:thread object-reference) ())
(define-class thread-group-mirror (jdi:thread-group object-reference)
  (jdi:parent))

(defmethod print-object ((object object-reference) stream)
  (print-unreadable-object (object stream :type nil :identity nil)
    (format stream "~A #~D"
	    (jdi:jdwp-handler-case
                (jdi:name (jdi:reference-type object))
              (:INVALID_OBJECT ()
                "broken mirror"))
	    (id object))))

(defmethod print-object ((object thread-mirror) stream)
  (print-unreadable-object (object stream :type t :identity nil)
    (format stream "~A #~D" (jdi:name object) (id object))))

(defmethod print-object ((object thread-group-mirror) stream)
  (print-unreadable-object (object stream :type t :identity nil)
    (format stream "~A #~D" (jdi:name object) (id object))))

(defun make-reference (class vm id &rest initargs)
  (if (zerop id)
      nil
      (apply #'make-instance class
	     :vm vm
	     :id id
	     initargs)))

(define-class frame-mirror (jdi:frame mirror)
  (jdi:parent thread))

(define-class local-mirror (jdi:local-variable mirror) (jdi:argumentp))

(defmethod print-object ((object local-mirror) stream)
  (print-unreadable-object (object stream :type t :identity nil)
    (format stream "~A~@[ ~A~] @~D"
	    (jdi:signature object)
	    (jdi:name object)
	    (id object))))



;;;; Misc

(defun jdi:jdwp-connect (host port)
  (when (stringp host)
    (setf host (get-host-by-name host)))
  (let ((stream (socket-stream (make-socket host port)
                               :element-type '(unsigned-byte 8))))
    (unless (shake-hands-with-server stream)
      (close stream)
      (error "Handshake error"))
    (make-instance 'client-vm :connection-stream stream)))

(defun shake-hands-with-server (s)
  (write-sequence *handshake* s)
  (force-output s)
  (let ((x (make-bytes (length *handshake*))))
    (read-sequence x s)
    (equalp x *handshake*)))

#+(or)
(defun read-event-body (vm)
  (subseq (in-buffer vm) (in-ptr vm)))

(defun fetch-events (vm)
  (while (listen (connection-stream vm))
    (fetch-message vm)
    (let* ((id (wp-read-int vm))
	   (flags (wp-read-byte vm))
	   (x (wp-read-short vm)))
      (declare (ignore id))
      (assert (zerop flags))
      (assert (eql x #x4064))
      (queue-insert (event-queue vm) (read-event-set vm)))))

(defun fetch-reply (vm expected-id)
  (loop
    (fetch-message vm)
    (let* ((id (wp-read-int vm))
	   (flags (wp-read-byte vm))
	   (x (wp-read-short vm)))
      (when (eql flags #x80)
	(unless (zerop x)
	  (error 'jdi:jdwp-error :code x))
	(assert (eql id expected-id))	;XXX do we need async. execution?
	(return))
      (assert (eql x #x4064))
      (when
	  ;; We cannot decode events before ID sizes are known.  But
	  ;; we haven't requested events at that point anyway, so this
	  ;; should only be a VM start event.  Let's ignore it.
	  (slot-boundp vm 'field-id-size)
	(queue-insert (event-queue vm) (read-event-set vm))))))

(defun event-kind-to-class (kind)
  (ecase kind
    (#.+single-step+	'jdi:step-event)
    (#.+breakpoint+	'jdi:breakpoint-event)
    ;; (#.+frame-pop+	'jdi:-event)
    (#.+exception+	'jdi:exception-event)
    ;; (#.+user-defined+	'jdi:-event)
    (#.+thread-start+	'jdi:thread-start-event)
    (#.+thread-death+	'jdi:thread-death-event)
    (#.+class-prepare+	'jdi:class-prepare-event)
    (#.+class-unload+	'jdi:class-unload-event)
    ;; ?
    ;; (#.+class-load+	'jdi:-event)
    (#.+field-access+	'jdi:field-access-event)
    (#.+field-modification+ 'jdi:field-modification-event)
    ;; (#.+exception-catch+	'jdi:-event)
    (#.+method-entry+	'jdi:method-entry-event)
    (#.+method-exit+	'jdi:method-exit-event)
    (#.+vm-start+		'jdi:vm-start-event)
    (#.+vm-death+		'jdi:vm-death-event)))

(defun read-event-set (vm)
  (let ((suspend-policy
	 (ecase (wp-read-byte vm)
	   (0 nil)
	   (1 :thread)
	   (2 t)))
	(n (wp-read-int vm)))
    (with-collector (collect :result events)
      (dotimes (x n)
	(let* ((kind (wp-read-byte vm))
	       (id (wp-read-int vm))
	       (request (find id (event-requests vm) :key #'request-id))
	       (event (make-instance (event-kind-to-class kind)
			:request request)))
	  (assert request)
	  (unless (member kind (list +class-unload+ +vm-death+))
	    (setf (jdi:thread event)
		  (make-reference 'thread-mirror vm (wp-read-object vm))))
	  (collect event)
	  (ecase kind
	    (#.+vm-start+)
	    ((#.+single-step+ #.+breakpoint+ #.+method-entry+ #.+method-exit+)
	      (setf (jdi:location event) (read-location vm)))
	    (#.+exception+
	      (setf (jdi:location event) (read-location vm))
	      (setf (jdi:exception event) (read-value vm "L"))
	      (setf (jdi:catch-location event) (read-location vm)))
	    ((#.+thread-start+ #.+thread-death+))
	    (#.+class-prepare+
	      (let* ((tag (wp-read-byte vm))
		     (id (wp-read-type vm))
		     (signature (wp-read-string vm))
		     (status (wp-read-int vm))
		     (class (make-type tag vm id signature)))
		(declare (ignore status)) ;XXX
		(pushnew class (cached-classes vm) :key #'id)
		(setf (jdi:class event) class)))
	    (#.+class-unload+
	      ;; `greatly limited' indeed -- wir braeuchten entweder die
	      ;; id oder den classloader, um die Klasse identifizieren zu 
	      ;; koennen.
	      (setf (jdi:signature event) (wp-read-string vm)))
	    ((#.+field-access+ #.+field-modification+)
	      ;; XXX eine Location bezieht sich doch schon auf einen Typ?
	      (setf (jdi:location event) (read-location vm))
	      (wp-read-byte vm)
	      (let* ((type (wp-read-type vm))
		     (class (find-class-mirror type vm))
		     (field (wp-read-field vm)))
		;; (setf (jdi:class event) class)
		(setf (jdi:field event)
		      (find field (jdi:fields class) :key #'id)))
	      (setf (jdi:object event) (read-value vm "L"))
	      (when (eq kind +field-modification+)
		(setf (jdi:value vm) (read-value vm))))
	    (#.+vm-death+
	      ;; nada
	      ))))
      (make-instance 'jdi:event-set
	:suspend-policy suspend-policy
	:events events))))

(defmethod jdi:read-event-set ((vm client-vm) &key wait)
  (fetch-events vm)
  (when wait
    (until (queue-listen (event-queue vm))
      (dlisp::wait-for-input-available (connection-stream vm))
      (fetch-events vm)))
  (queue-delete (event-queue vm)))

(defun finish-command (c set command &optional (id (n++ (next-request-id c))))
  (let ((buf (out-buffer c))
	(s (connection-stream c)))
    (let ((length (fill-pointer buf)))
      (setf (fill-pointer buf) 0)
      (wp-write-int length c)
      (wp-write-int id c)
      (wp-write-byte 0 c)
      (wp-write-byte set c)
      (wp-write-byte command c)
      (setf (fill-pointer buf) length))
    (write-sequence buf s)
    (force-output s))
  id)

(defun read-value (vm &optional expected-signature)
  (let ((tag (code-char (wp-read-byte vm))))
    (when expected-signature
      (let ((expected-tag (char expected-signature 0)))
	(assert (or (and (char= expected-tag #\L)
			 (member tag '(#\L #\s #\t #\g #\l #\c)))
		    (eql tag expected-tag)))))
    (read-untagged-value vm tag)))

(defun read-array-region (vm)
  (let* ((tag (code-char (wp-read-byte vm)))
	 (objectp (member tag '(#\L #\s #\t #\g #\l #\c)))
	 (length (wp-read-int vm))
	 (result (make-array length)))
    (dotimes (i length)
      (setf (elt result i)
	    (if objectp (read-value vm) (read-untagged-value vm tag))))
    result))

(defun read-untagged-value (vm tag)
  (case tag
    (#\B (wp-read-byte vm))
    (#\C (let ((c (wp-read-short vm))) (or (code-char c) c)))
    (#\S (wp-read-short vm))
    (#\I (wp-read-int vm))
    (#\J (wp-read-long vm))
    (#\F (|java/lang/Float.intBitsToFloat(I)| (wp-read-int vm)))
    (#\D (|java/lang/Double.longBitsToDouble(J)| (wp-read-long vm)))
    (#\Z (if (wp-read-byte vm) :true :false))
    (t
      (make-reference (ecase tag
			(#\[ 'array-reference)
			(#\t 'thread-mirror)
			(#\g 'thread-group-mirror)
			((#\L #\s #\l #\c) 'object-reference))
	  vm
	  (wp-read-object vm)))))



;;;; VIRTUAL-MACHINE

(defun fetch-vm-description (vm)
  (reset-out-buffer vm)
  (fetch-reply vm (finish-command vm 1 1))
  (setf (jdi:description vm) (wp-read-string vm))
  (setf (jdi:jdwp-version vm) (cons (wp-read-int vm) (wp-read-int vm)))
  (setf (jdi:version vm) (wp-read-string vm))
  (setf (jdi:name vm) (wp-read-string vm)))

(defun make-type (tag vm id signature)
  (make-instance
      (case tag
	(#.+typetag-array+ 'array-type-mirror)
	(#.+typetag-class+ 'class-type-mirror)
	(#.+typetag-interface+ 'interface-type-mirror))
    :vm vm
    :id id
    :signature signature))

(defmethod jdi:all-classes ((vm client-vm))
  (unless (slot-boundp vm 'cached-classes)
    (fetch-reply vm (finish-command vm 1 3))
    (with-collector (collect :result classes)
      (dotimes (x (wp-read-int vm))
	(let ((type-tag (wp-read-byte vm))
	      (type-id (wp-read-type vm))
	      (signature (wp-read-string vm))
	      (status (wp-read-int vm)))
	  (declare (ignore status))	;XXX
	  (collect
	   (make-type type-tag vm type-id signature)))) 
      (setf (cached-classes vm) classes)))
  (cached-classes vm))

(defmethod jdi:all-threads ((vm client-vm))
  (reset-out-buffer vm)
  (fetch-reply vm (finish-command vm 1 4))
  (with-collector ()
    (dotimes (x (wp-read-int vm))
      (collect (make-reference 'thread-mirror vm (wp-read-object vm))))))

(defmethod jdi:dispose ((vm client-vm))
  (reset-out-buffer vm)
  (finish-command vm 1 6))

(defmethod jdi:suspend ((vm client-vm))
  (reset-out-buffer vm)
  (fetch-reply vm (finish-command vm 1 8)))

(defmethod jdi:resume ((vm client-vm))
  (reset-out-buffer vm)
  (fetch-reply vm (finish-command vm 1 9)))

(defun fetch-id-sizes (vm)
  (reset-out-buffer vm)
  (fetch-reply vm (finish-command vm 1 7))
  (setf (field-id-size vm) (wp-read-int vm))
  (setf (method-id-size vm) (wp-read-int vm))
  (setf (object-id-size vm) (wp-read-int vm))
  (setf (type-id-size vm) (wp-read-int vm))
  (setf (frame-id-size vm) (wp-read-int vm)))

(defun version<= (a b)
  (if (= (car a) (car b))
      (<= (cdr a) (cdr b))
      (< (car a) (car b))))

(defun fetch-capabilities (vm)
  (setf (capabilities vm)
	(let* ((new (version<= (cons 1 4) (jdi:jdwp-version vm)))
	       (v (make-array (if new 31 7))))
	  (reset-out-buffer vm)
	  (fetch-reply vm (finish-command vm 1 (if new 17 12)))
	  (dotimes (i (length v))
	    (setf (elt v i) (wp-read-boolean vm)))
	  v)))

(defmethod jdi:get-capability ((vm client-vm) name)
  (let ((id (position name +capability-names+)))
    (unless id
      (error "not a valid capability name: ~D" name))
    (if (< id (length (capabilities vm)))
	(elt (capabilities vm) id)
	nil)))

(defmethod jdi:top-level-thread-groups ((vm client-vm))
  (reset-out-buffer vm)
  (fetch-reply vm (finish-command vm 1 5))
  (with-collector ()
    (dotimes (x (wp-read-int vm))
      (collect (make-reference 'thread-group-mirror
		   vm
		   (wp-read-object vm)
		 ;; Hmm, die Sun-VM mag es nicht, wenn man das Parent
		 ;; einer Top-Level Thread-Group abfragt?
		 :parent nil)))))

(defmethod jdi:base-directory ((vm client-vm))
  (ensure-class-paths vm)
  (base-directory vm))

(defmethod jdi:class-path ((vm client-vm))
  (ensure-class-paths vm)
  (class-path vm))

(defmethod jdi:bootstrap-class-path ((vm client-vm))
  (ensure-class-paths vm)
  (bootstrap-class-path vm))

(defun ensure-class-paths (vm)
  (unless (slot-boundp vm 'class-path)
    (reset-out-buffer vm)
    (fetch-reply vm (finish-command vm 1 13))
    (setf (base-directory vm) (wp-read-string vm))
    (flet ((read-strings ()
	     (with-collector ()
	       (dotimes (x (wp-read-int vm))
		 (collect (wp-read-string vm))))))
      (setf (class-path vm) (read-strings))
      (setf (bootstrap-class-path vm) (read-strings)))))

(defmethod jdi:classes-by-signature ((vm client-vm) signature)
  (flet ((select-from (classes)
	   (remove signature classes :key #'jdi:signature :test-not #'equal)))
    (or (select-from (primitive-classes vm))
	(select-from (cached-classes vm)))))



;;;; REFERENCE-TYPE

(defmethod jdi:signature ((class reference-type-mirror))
  (unless (slot-boundp class 'signature)
    (let ((vm (vm class)))
      (reset-out-buffer vm)
      (wp-write-type (id class) vm)
      (fetch-reply vm (finish-command vm 2 1))
      (setf (signature class) (wp-read-string vm))))
  (signature class))

(defmethod jdi:fields ((class reference-type-mirror))
  (members class 4 'field-mirror 'fields #'wp-read-field))

(defmethod jdi:methods ((class reference-type-mirror))
  (members class 5 'method-mirror 'methods #'wp-read-method))

(defun collect-inherited (fn class)
  (let ((result (funcall fn class))
	(super (typecase class
		 (jdi:interface-type nil)
		 (t (jdi:superclass class)))))
    (when super
      (setf result (append result (collect-inherited fn super))))
    (dolist (i (jdi:interfaces class))
      (setf result (append result (collect-inherited fn i))))
    result))

(defmethod jdi:all-fields ((class reference-type-mirror))
  (collect-inherited #'jdi:fields class))

(defun method-equal (a b)
  (and (equal (jdi:name a) (jdi:name b))
       (equal (jdi:signature a) (jdi:signature b))))

(defmethod jdi:all-methods
    ((class reference-type-mirror) &key exclude-overriden)
  (let ((result (collect-inherited #'jdi:methods class)))
    (if exclude-overriden
	(remove-duplicates result :test #'method-equal :from-end t)
	result)))

(defun members (type command lisp-class slot id-reader)
  (unless (slot-boundp type slot)
    (let ((vm (vm type)))
      (reset-out-buffer vm)
      (wp-write-type (id type) vm)
      (fetch-reply vm (finish-command vm 2 command))
      (setf (slot-value type slot)
	    (with-collector ()
	      (dotimes (x (wp-read-int vm))
		(let ((id (funcall id-reader vm))
		      (name (wp-read-string vm))
		      (signature (wp-read-string vm))
		      (modifiers (wp-read-int vm)))
		  (collect
		   (make-instance lisp-class
		     :vm vm
		     :id id
		     :declaring-type type
		     :name name
		     :modifiers modifiers
		     :signature signature))))))))
  (slot-value type slot))

(defun find-class-mirror (id vm)
  (if (zerop id)
      nil
      (let ((c (find id (jdi:all-classes vm) :key #'id)))
	(assert c)
	c)))

(defmethod jdi:field-value (field (class reference-type-mirror))
  (car (jdi:field-values (list field) class)))

(defmethod jdi:field-values (fields (class reference-type-mirror))
  (let ((vm (vm class)))
    (reset-out-buffer vm)
    (wp-write-type (id class) vm)
    (wp-write-int (length fields) vm)
    (dolist (field fields)
      (wp-write-field (id field) vm))
    (fetch-reply vm (finish-command vm 2 6))
    (with-collector ()
      (dotimes (x (wp-read-int vm))
	(collect (read-value vm))))))

(defmethod jdi:interfaces ((class reference-type-mirror))
  (if (slot-boundp class 'cached-interfaces)
      (slot-value class 'cached-interfaces)
      (setf (slot-value class 'cached-interfaces)
	    (let ((vm (vm class)))
	      (reset-out-buffer vm)
	      (wp-write-type (id class) vm)
	      (fetch-reply vm (finish-command vm 2 10))
	      (with-collector ()
		(dotimes (x (wp-read-int vm))
		  (collect (find-class-mirror (wp-read-type vm) vm))))))))

(defmethod jdi:source-file ((class reference-type-mirror))
  (let ((vm (vm class)))
    (reset-out-buffer vm)
    (wp-write-type (id class) vm)
    (fetch-reply vm (finish-command vm 2 7))
    (wp-read-string vm)))



;;;; CLASS-TYPE

(defmethod %id-to-signature (vm id)
  (reset-out-buffer vm)
  (wp-write-type id vm)
  (fetch-reply vm (finish-command vm 2 1))
  (wp-read-string vm))

(defmethod jdi:superclass ((class reference-type-mirror))
  (if (slot-boundp class 'cached-superclass)
      (slot-value class 'cached-superclass)
      (setf (slot-value class 'cached-superclass)
	    (let ((vm (vm class)))
	      (reset-out-buffer vm)
	      (wp-write-type (id class) vm)
	      (fetch-reply vm (finish-command vm 3 1))
	      (find-class-mirror (wp-read-type vm) vm)))))

(defmethod jdi:subclasses ((class class-type-mirror))
  (remove-if-not (lambda (c)
		   (and (typep c 'jdi:class-type)
			(eq (jdi:superclass c) class)))
		 (jdi:all-classes (vm class))))



;;;; INTERFACE-TYPE

(defmethod jdi:subinterfaces ((class interface-type-mirror))
  (remove-if-not (lambda (c)
		   (and (typep c 'jdi:interface-type)
			(member class (jdi:interfaces c))))
		 (jdi:all-classes (vm class))))

(defmethod jdi:superinterfaces ((class interface-type-mirror))
  (jdi:interfaces class))



;;;; ARRAY-TYPE

(defmethod jdi:component-signature ((class array-type-mirror))
  (subseq (jdi:signature class) 1))

(defmethod jdi:component-type ((class array-type-mirror))
  (find (jdi:component-signature class)
	(jdi:all-classes (vm class))
	:test #'equal
	:key #'jdi:signature))

(defmethod jdi:component-type-name ((class array-type-mirror))
  (jdi:name (jdi:component-type class)))



;;;; METHOD

(defmethod jdi:lines ((method method-mirror))
  (let ((vm (vm method))
	(class (jdi:declaring-type method)))
    (reset-out-buffer vm)
    (wp-write-type (id class) vm)
    (wp-write-method (id method) vm)
    (fetch-reply vm (finish-command vm 6 1))
    (let ((start (wp-read-long vm))
	  (end (wp-read-long vm)))
      (declare (ignore start end))	;XXX
      (with-collector ()
	(dotimes (x (wp-read-int vm))
	  (let ((code-index (wp-read-long vm))
		(line (wp-read-int vm)))
	    (collect
	     (make-instance 'jdi:location
	       :type class
	       :code-index code-index
	       :method-id (id method)
	       :line-number line))))))))

(defmethod jdi:location ((method method-mirror))
  (car (jdi:lines method)))

(defmethod jdi:local-variables ((method method-mirror))
  (let ((vm (vm method)))
    (reset-out-buffer vm)
    (wp-write-type (id (jdi:declaring-type method)) vm)
    (wp-write-method (id method) vm)
    (fetch-reply vm (finish-command vm 6 2))
    (with-collector ()
      (let ((nargs (wp-read-int vm))
	    (nslots (wp-read-int vm)))
	(dotimes (n nslots)
	  (let ((code-index (wp-read-long vm))
		(name (wp-read-string vm))
		(signature (wp-read-string vm))
		(length (wp-read-int vm))
		(slot (wp-read-int vm)))
	    (declare (ignore code-index length))
	    (collect
	     (make-instance 'local-mirror
	       :vm vm
	       :id slot
	       :name name
	       :argumentp (< n nargs)
	       :signature signature))))))))

(defun signature-word-length (str)
  (if (or (equal str "D") (equal str "L")) 2 1))

(defmethod jdi:guess-local-variables ((method method-mirror) &optional class)
  (unless class
    (setf class (jdi:declaring-type method)))
  (let ((vm (vm method))
	(i 0)
	(args (jdi:argument-type-names method)))
    (unless (jdi:staticp method)
      (push (jdi:signature class) args))
    (with-collector ()
      (dolist (arg args)
	(collect
	 (make-instance 'local-mirror
	   :vm vm
	   :id i
	   :name nil
	   :signature arg))
	(incf i (signature-word-length arg))))))



;;;; OBJECT-REFERENCE

(defmethod jdi:reference-type ((object object-reference))
  (let ((vm (vm object)))
    (reset-out-buffer vm)
    (wp-write-object (id object) vm)
    (fetch-reply vm (finish-command vm 9 1))
    (let ((tag (wp-read-byte vm))
	  (id (wp-read-type vm)))
      (declare (ignore tag))
      (find-class-mirror id vm))))

(defmethod jdi:field-value (field (object object-reference))
  (car (jdi:field-values (list field) object)))

(defmethod jdi:field-values (fields (object object-reference))
  (let ((vm (vm object)))
    (reset-out-buffer vm)
    (wp-write-object (id object) vm)
    (wp-write-int (length fields) vm)
    (dolist (field fields)
      (wp-write-field (id field) vm))
    (fetch-reply vm (finish-command vm 9 2))
    (with-collector ()
      (dotimes (x (wp-read-int vm))
	(collect (read-value vm))))))


;;;; THREAD

(defmethod jdi:name :around ((thread thread-mirror))
  (if (slot-boundp thread 'jdi:name)
      (call-next-method)
      (setf (slot-value thread 'jdi:name)
	    (let ((vm (vm thread)))
	      (reset-out-buffer vm)
	      (wp-write-object (id thread) vm)
	      (fetch-reply vm (finish-command vm 11 1))
	      (wp-read-string vm)))))

(defmethod jdi:suspend ((thread thread-mirror))
  (let ((vm (vm thread)))
    (reset-out-buffer vm)
    (wp-write-object (id thread) vm)
    (fetch-reply vm (finish-command vm 11 2))))

(defmethod jdi:resume ((thread thread-mirror))
  (let ((vm (vm thread)))
    (reset-out-buffer vm)
    (wp-write-object (id thread) vm)
    (fetch-reply vm (finish-command vm 11 3))))

(defmethod jdi:status ((thread thread-mirror))
  (nth-value 0 (status thread)))

(defmethod jdi:suspendedp ((thread thread-mirror))
  (nth-value 1 (status thread)))

(defmethod status ((thread thread-mirror))
  (let ((vm (vm thread)))
    (reset-out-buffer vm)
    (wp-write-object (id thread) vm)
    (fetch-reply vm (finish-command vm 11 4))
    (values
     (ecase (wp-read-int vm)
       (#xffffffff NIL)			;?!
       (0 :zombie)
       (1 :running)
       (2 :sleeping)
       (3 :monitor)
       (4 :wait))
     (ecase (wp-read-int vm)
       (0 nil)
       (1 t)))))

(defmethod jdi:group ((thread thread-mirror))
  (let ((vm (vm thread)))
    (reset-out-buffer vm)
    (wp-write-object (id thread) vm)
    (fetch-reply vm (finish-command vm 11 5))
    (make-reference 'thread-group-mirror vm (wp-read-object vm))))

(defun read-location (vm)
  (let ((type-tag (wp-read-byte vm))
	(class-id (wp-read-type vm))
	(method (wp-read-method vm))
	(index (wp-read-long vm)))
    (declare (ignore type-tag))
    (let ((class (find-class-mirror class-id vm)))
      (unless class
	(error "class ~A not found" (%id-to-signature vm class)))
      (make-instance 'jdi:location
	:code-index index
	:type class
	:method-id method))))

(defun write-location (location vm)
  (wp-write-byte (class-tag (jdi:type location)) vm)
  (wp-write-type (id (jdi:type location)) vm)
  (wp-write-method (id (jdi:method location)) vm)
  (wp-write-long (jdi:code-index location) vm))

(defmethod jdi:method ((location jdi:location))
  (find (method-id location) (jdi:methods (jdi:type location)) :key #'id))

(defmethod jdi:frames ((thread thread-mirror))
  (let ((vm (vm thread)))
    (reset-out-buffer vm)
    (wp-write-object (id thread) vm)
    (wp-write-int 0 vm)
    (wp-write-int -1 vm)
    (fetch-reply vm (finish-command vm 11 6))
    (with-collector ()
      (dotimes (x (wp-read-int vm))
	(collect
	 (make-instance 'frame-mirror
	   :vm vm
	   :thread thread
	   :id (wp-read-frame vm)
	   :location (read-location vm)))))))



;;;; THREAD-GROUP

(defmethod jdi:name :around ((thread-group thread-group-mirror))
  (if (slot-boundp thread-group 'jdi:name)
      (call-next-method)
      (setf (slot-value thread-group 'jdi:name)
	    (let ((vm (vm thread-group)))
	      (reset-out-buffer vm)
	      (wp-write-object (id thread-group) vm)
	      (fetch-reply vm (finish-command vm 12 1))
	      (wp-read-string vm)))))

(defmethod jdi:parent :around ((thread-group thread-group-mirror))
  (if (slot-boundp thread-group 'jdi:parent)
      (call-next-method)
      (setf (slot-value thread-group 'jdi:parent)
	    (let ((vm (vm thread-group)))
	      (reset-out-buffer vm)
	      (wp-write-object (id thread-group) vm)
	      (fetch-reply vm (finish-command vm 12 2))
	      (make-reference 'thread-group-mirror vm (wp-read-object vm))))))

(defmethod jdi:child-threads ((group thread-group-mirror))
  (nth-value 0 (children group)))

(defmethod jdi:child-groups ((group thread-group-mirror))
  (nth-value 1 (children group)))

(defmethod children ((group thread-group-mirror))
  (let ((vm (vm group)))
    (reset-out-buffer vm)
    (wp-write-object (id group) vm)
    (fetch-reply vm (finish-command vm 12 3))
    (values
     (with-collector ()
       (dotimes (x (wp-read-int vm))
	 (collect (make-reference 'thread-mirror vm (wp-read-object vm)))))
     (with-collector ()
       (dotimes (x (wp-read-int vm))
	 (collect (make-reference 'thread-group-mirror
		      vm
		      (wp-read-object vm))))))))



;;; ARRAY-REFERENCE

(defmethod jdi:length ((array array-reference))
  (let ((vm (vm array)))
    (reset-out-buffer vm)
    (wp-write-object (id array) vm)
    (fetch-reply vm (finish-command vm 13 1))
    (wp-read-int vm)))

(defmethod jdi:elt ((array array-reference) i)
  (elt (jdi:data array i (1+ i)) 0))

(defmethod jdi:data
    ((array array-reference) &optional (start 0) (end (jdi:length array)))
  (if (< start end)
      (let ((vm (vm array)))
	(reset-out-buffer vm)
	(wp-write-object (id array) vm)
	(wp-write-int 0 vm)
	(wp-write-int (- end start) vm)
	(fetch-reply vm (finish-command vm 13 2))
	(read-array-region vm))
      #()))



;;; LOCAL-VARIABLE

(defmethod jdi:type ((local local-mirror))
  (find (jdi:signature local)
	(jdi:all-classes (vm local))
	:key #'jdi:signature
	:test #'equal))



;;; EVENT and EVENT-REQUEST

(defun request-kind (request)
  ;; (jdi:-request	+frame-pop+)
  ;; (jdi:-request	+user-defined+)
  ;; ?
  ;; (jdi:-request	+class-load+)
  ;; (jdi:-request	+exception-catch+)
  (etypecase request
    (jdi:step-request		+single-step+)
    (jdi:breakpoint-request	+breakpoint+)
    (jdi:exception-request	+exception+)
    (jdi:thread-start-request	+thread-start+)
    (jdi:thread-death-request	+thread-death+)
    (jdi:class-prepare-request	+class-prepare+)
    (jdi:class-unload-request	+class-unload+)
    (jdi:field-access-request	+field-access+)
    (jdi:field-modification-request +field-modification+)
    (jdi:method-entry-request	+method-entry+)
    (jdi:method-exit-request	+method-exit+)
    (jdi:vm-start-request	+vm-start+)
    (jdi:vm-death-request	+vm-death+)))

(defmethod jdi:request-event ((vm client-vm) request)
  (reset-out-buffer vm)
  (wp-write-byte (request-kind request) vm)
  (let ((suspend-policy
	 (ecase (jdi:suspend-policy request)
	   ((nil) 0)
	   (:thread 1)
	   ((t) 2))))
    (wp-write-byte suspend-policy vm))
  (wp-write-int (length (modifiers request)) vm)
  (dolist (modifier (modifiers request))
    (write-modifier modifier vm))
  (fetch-reply vm (finish-command vm 15 1))
  (setf (request-id request) (wp-read-int vm))
  (setf (vm request) vm)
  (push request (event-requests vm))
  request)

(defmethod jdi:resume ((set jdi:event-set))
  (let ((e (car (jdi:events set))))
    (ecase (jdi:suspend-policy set)
      ((nil))
      ((t) (jdi:resume (vm e)))
      (:thread (jdi:resume (jdi:thread e))))))

(defmethod jdi:cancel-request ((vm client-vm) request)
  (reset-out-buffer vm)
  (wp-write-byte (request-kind request) vm)
  (wp-write-int (request-id request) vm)
  (fetch-reply vm (finish-command vm 15 2))
  nil)

(defmethod write-modifier ((modifier location-modifier) vm)
  (wp-write-byte 7 vm)
  (write-location (location modifier) vm))

(defmethod write-modifier ((modifier exception-only-modifier) vm)
  (wp-write-byte 8 vm)
  (wp-write-type (id (reference-type modifier)) vm)
  (wp-write-boolean (caughtp modifier) vm)
  (wp-write-boolean (uncaughtp modifier) vm))

(defmethod write-modifier ((modifier field-modifier) vm)
  (wp-write-byte 9 vm)
  (wp-write-type (id (jdi:declaring-type (field modifier))) vm)
  (wp-write-field (id (field modifier)) vm))

(defmethod write-modifier ((modifier step-modifier) vm)
  (wp-write-byte 10 vm)
  (wp-write-object (id (thread modifier)) vm)
  (wp-write-int (size modifier) vm)
  (wp-write-int (depth modifier) vm))



;;; FRAME

(defmethod jdi:get-values ((frame frame-mirror) locals)
  (let ((vm (vm frame)))
    (reset-out-buffer vm)
    (wp-write-object (id (thread frame)) vm)
    (wp-write-frame (id frame) vm)
    (wp-write-int (length locals) vm)
    (dolist (local locals)
      (wp-write-int (id local) vm)
      (wp-write-byte (char-code (char (jdi:signature local) 0)) vm))
    (fetch-reply vm (finish-command vm 16 1))
    (assert (eql (wp-read-int vm) (length locals)))
    (with-collector ()
      (dolist (local locals)
	(collect (read-value vm (jdi:signature local)))))))

(defmethod jdi:this-object ((frame frame-mirror))
  (let ((vm (vm frame)))
    (reset-out-buffer vm)
    (wp-write-object (id (thread frame)) vm)
    (wp-write-frame (id frame) vm)
    (fetch-reply vm (finish-command vm 16 3))
    (read-value vm "L")))
