;;;; Copyright (c) 2004 David Lichteblau <david@lichteblau.com>
;;;;
;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :jdii)

(defun jni-to-java (str)
  (let* ((narray (position #\[ str :test-not #'char=))
	 (relative-name (subseq str narray)))
    (with-output-to-string (result)
      (write-string
       (cond
	 ((char= (char relative-name 0) #\L)
	   (let ((end (1- (length relative-name))))
	     (assert (char= (char relative-name end) #\;))
	     (substitute #\. #\/ (subseq relative-name 1 end))))
	 ((equal relative-name "B") "byte")
	 ((equal relative-name "C") "char")
	 ((equal relative-name "D") "double")
	 ((equal relative-name "F") "float")
	 ((equal relative-name "I") "int")
	 ((equal relative-name "J") "long")
	 ((equal relative-name "S") "short")
	 ((equal relative-name "V") "void")
	 ((equal relative-name "Z") "boolean")
	 (t
	   (error "signature parse error")))
       result)
      (dotimes (x narray)
	(write-string "[]" result)))))

(defun read-field-descriptor (in)
  (with-output-to-string (out)
    (labels ((doit ()
	       (let ((c (read-char in)))
		 (write-char c out)
		 (case c
		   ((#\B #\C #\D #\F #\I #\J #\S #\Z))
		   (#\V)
		   (#\[ (doit))
		   (#\L (loop
			    for c = (read-char in)
			    do (write-char c out)
			    until (char= c #\;)))
		   (t
		     (unread-char c in)
		     (return-from read-field-descriptor nil))))))
      (doit))))

(defun parse-method-descriptor (str)
  (with-input-from-string (in str)
    (assert (char= (read-char in) #\())
    (let ((arguments
	   (loop for a = (read-field-descriptor in) while a collect a)))
      (assert (char= (read-char in) #\)))
      (values arguments
	      (read-field-descriptor in)))))

(defconstant +acc_public+ #x0001)
(defconstant +acc_private+ #x0002)
(defconstant +acc_protected+ #x0004)
(defconstant +acc_static+ #x0008)
(defconstant +acc_final+ #x0010)
(defconstant +acc_synchronized+ #x0020)
(defconstant +acc_super+ #x0020)
(defconstant +acc_volatile+ #x0040)
(defconstant +acc_transient+ #x0080)
(defconstant +acc_native+ #x0100)
(defconstant +acc_interface+ #x0200)
(defconstant +acc_abstract+ #x0400)
(defconstant +acc_strict+ #x0800)

(defmethod jdi:package-private-p ((member jdi:member))
  (zerop (logand (logior +acc_private+ +acc_protected+ +acc_public+)
		 (jdi:modifiers member))))

(defmethod jdi:privatep ((member jdi:member))
  (logtest +acc_private+ (jdi:modifiers member)))

(defmethod jdi:protectedp ((member jdi:member))
  (logtest +acc_protected+ (jdi:modifiers member)))

(defmethod jdi:publicp ((member jdi:member))
  (logtest +acc_public+ (jdi:modifiers member)))

(defmethod jdi:finalp ((member jdi:member))
  (logtest +acc_final+ (jdi:modifiers member)))

(defmethod jdi:staticp ((member jdi:member))
  (logtest +acc_static+ (jdi:modifiers member)))

(defmethod jdi:syntheticp ((member jdi:member))
  (error "not implemented"))

(defmethod jdi:transientp ((member jdi:field))
  (logtest +acc_transient+ (jdi:modifiers member)))

(defmethod jdi:volatilep ((member jdi:field))
  (logtest +acc_volatile+ (jdi:modifiers member)))

(defmethod jdi:abstractp ((member jdi:method))
  (logtest +acc_abstract+ (jdi:modifiers member)))

(defmethod jdi:nativep ((member jdi:method))
  (logtest +acc_native+ (jdi:modifiers member)))

(defmethod jdi:synchronizedp ((member jdi:method))
  (logtest +acc_synchronized+ (jdi:modifiers member)))

(defparameter +capability-names+
    #(:CAN-WATCH-FIELD-MODIFICATION
      :CAN-WATCH-FIELD-ACCESS
      :CAN-GET-BYTECODES
      :CAN-GET-SYNTHETIC-ATTRIBUTE
      :CAN-GET-OWNED-MONITOR-INFO
      :CAN-GET-CURRENT-CONTENDED-MONITOR
      :CAN-GET-MONITOR-INFO
      :CAN-REDEFINE-CLASSES
      :CAN-ADD-METHOD
      :CAN-UNRESTRICTEDLY-REDEFINE-CLASSES
      :CAN-POP-FRAMES
      :CAN-USE-INSTANCE-FILTERS
      :CAN-GET-SOURCE-DEBUG-EXTENSION

      :CAN-REQUEST-VM-DEATH-EVENT
      :CAN-SET-DEFAULT-STRATUM))

;;;; JDWP-ERROR

(define-condition jdi:jdwp-error ()
    ((code :initarg :code :accessor jdi:jdwp-error-code))
  (:report
   (lambda (condition stream)
     (multiple-value-bind (key str)
	 (decode-error (jdi:jdwp-error-code condition))
       (format stream "JDWP Error ~A (~D): ~A"
	       key (jdi:jdwp-error-code condition) str)))))

(defmacro jdi:jdwp-handler-case (form &rest cases)
  (with-gensyms (condition)
    `(BLOCK JDWP-HANDLER-CASE
       (HANDLER-BIND
	((JDI:JDWP-ERROR
	  (LAMBDA (,condition)
		  (CASE (DECODE-ERROR (JDI:JDWP-ERROR-CODE ,condition))
			,@(with-collector ()
			    (dolist (case cases)
			      (destructuring-bind
				  (key (&optional c) &body body)
				  case
				(let ((bindings
				       (if c
					   `((,c ,condition))
					   '())))
				  (collect
				   `(,key
				     (RETURN-FROM JDWP-HANDLER-CASE
                                       (LET (,@bindings) ,@body))))))))))))
	,form))))
