;;;; Copyright (c) 2004 David Lichteblau <david@lichteblau.com>
;;;;
;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :jdii)

(defun make-buffer ()
  (make-array 256
	      :element-type '(unsigned-byte 8)
	      :fill-pointer 0
	      :adjustable t))

(define-class jdwp-connection ()
  (connection-stream
   (in-buffer :initform (make-buffer))
   in-ptr
   (out-buffer :initform (make-buffer))
   (next-request-id :initform 1)
   (next-object-id :initform 1)
   (ids :initform (make-hash-table))
   (objects :initform (make-hash-table))
   field-id-size
   method-id-size
   object-id-size
   type-id-size
   frame-id-size))

(defmethod print-object ((object jdwp-connection) stream)
  (print-unreadable-object (object stream :type t :identity t)
    (princ (connection-stream object) stream)))

(defun object->id (class connection)
  (if class
      (or (gethash class (ids connection))
	  (let ((id (n++ (next-object-id connection))))
	    (setf (gethash id (objects connection)) class)
	    (setf (gethash class (ids connection)) id)
	    id))
      0))

(defun id->object (id connection)
  (gethash id (objects connection)))

(defvar *handshake* #(74 68 87 80 45 72 97 110 100 115 104 97 107 101))

(defun wp-read-byte (c)
  (elt (in-buffer c) (n++ (in-ptr c))))

(defun wp-read-boolean (c)
  (/= (wp-read-byte c) 0))

(defun wp-read-short (c)
  (logior (ash (wp-read-byte c) 8)
	  (wp-read-byte c)))

(defun wp-read-int (c)
  (logior (ash (wp-read-byte c) 24)
	  (ash (wp-read-byte c) 16)
	  (ash (wp-read-byte c) 8)
	  (wp-read-byte c)))

(defun read-length (s)
  (logior (ash (read-byte s) 24)
	  (ash (read-byte s) 16)
	  (ash (read-byte s) 8)
	  (read-byte s)))

(defun wp-read-long (c)
  (logior (ash (wp-read-byte c) 56)
	  (ash (wp-read-byte c) 48)
	  (ash (wp-read-byte c) 40)
	  (ash (wp-read-byte c) 32)
	  (ash (wp-read-byte c) 24)
	  (ash (wp-read-byte c) 16)
	  (ash (wp-read-byte c) 8)
	  (wp-read-byte c)))

(defun wp-read-id (c length)
  (let ((result 0))
    (dotimes (x length)
      (setf result (logior (ash result 8) (wp-read-byte c))))
    result))

(defun wp-read-field  (c) (wp-read-id c (field-id-size c)))
(defun wp-read-method (c) (wp-read-id c (method-id-size c)))
(defun wp-read-object (c) (wp-read-id c (object-id-size c)))
(defun wp-read-type   (c) (wp-read-id c (type-id-size c)))
(defun wp-read-frame  (c) (wp-read-id c (frame-id-size c)))

(defun wp-read-string (c)
  (let* ((n (wp-read-int c))
	 (start (in-ptr c))
	 (end (+ (in-ptr c) n)))
    (setf (in-ptr c) end)
    (utf-8-to-string (subseq (in-buffer c) start end))))

#+(or)
(defun wp-read-location (c)
  (let* ((type-tag (wp-read-byte c))
	 (class-id (wp-read-type c))
	 (method-id (wp-read-method c))
	 (index (wp-read-long c)))
    (list type-tag class-id method-id index)))

(defun wp-write-byte (b c)
  (let ((buf (out-buffer c)))
    (vector-push-extend b buf (length buf))))

(defun wp-write-boolean (x c)
  (wp-write-byte (if x 1 0) c))

(defun wp-write-short (b c)
  (wp-write-byte (ldb (byte 8 08) b) c)
  (wp-write-byte (ldb (byte 8 00) b) c))

(defun wp-write-int (b c)
  (wp-write-byte (ldb (byte 8 24) b) c)
  (wp-write-byte (ldb (byte 8 16) b) c)
  (wp-write-byte (ldb (byte 8 08) b) c)
  (wp-write-byte (ldb (byte 8 00) b) c))

(defun wp-write-long (b c)
  (wp-write-byte (ldb (byte 8 56) b) c)
  (wp-write-byte (ldb (byte 8 48) b) c)
  (wp-write-byte (ldb (byte 8 40) b) c)
  (wp-write-byte (ldb (byte 8 32) b) c)
  (wp-write-byte (ldb (byte 8 24) b) c)
  (wp-write-byte (ldb (byte 8 16) b) c)
  (wp-write-byte (ldb (byte 8 08) b) c)
  (wp-write-byte (ldb (byte 8 00) b) c))

(defun wp-write-string (str c)
  (let* ((buf (out-buffer c))
	 (b (string-to-utf-8 str)))
    (wp-write-int (length b) c)
    (vector-push-sequence b buf)))

(defun wp-string-length (str)
  (+ 4 (length (string-to-utf-8 str))))

(defun wp-write-id (id c length)
  (for ((i :from (1- length) :downto 0))
    (wp-write-byte (logand (ash id (* i -8)) #xff) c)))

(defun wp-write-field  (x c) (wp-write-id x c (field-id-size c)))
(defun wp-write-method (x c) (wp-write-id x c (method-id-size c)))
(defun wp-write-object (x c) (wp-write-id x c (object-id-size c)))
(defun wp-write-type   (x c) (wp-write-id x c (type-id-size c)))
(defun wp-write-frame  (x c) (wp-write-id x c (frame-id-size c)))

(defun reset-out-buffer (c)
  (setf (fill-pointer (out-buffer c)) 11))

(defun fetch-message (c)
  (let* ((length (- (read-length (connection-stream c)) 4))
	 (buf (in-buffer c)))
    (adjust-vector-exponentially buf length t)
    (read-sequence buf (connection-stream c))
    (setf (in-ptr c) 0)))

(defun decode-error (code)
  (case code
    (0 (values :NONE "No error has occurred."))
    (10 (values :INVALID_THREAD "Passed thread is null, is not a valid thread or has exited."))
    (11 (values :INVALID_THREAD_GROUP "Thread group invalid."))
    (12 (values :INVALID_PRIORITY "Invalid priority."))
    (13 (values :THREAD_NOT_SUSPENDED "The specified thread has not been suspended by an event."))
    (14 (values :THREAD_SUSPENDED "Thread already suspended."))
    (20
      ;; "This reference type has been unloaded and garbage collected."
      ;; -- Das ist doch Unsinn.  INVALID_OBJECT bezieht sich doch offenbar
      ;; auf ein Objekt, die Beschreibung aber auf eine Klasse.  Das passt
      ;; nicht zusammen.
      (values :INVALID_OBJECT ""))
    (21 (values :INVALID_CLASS "Invalid class."))
    (22 (values :CLASS_NOT_PREPARED "Class has been loaded but not yet prepared."))
    (23 (values :INVALID_METHODID "Invalid method."))
    (24 (values :INVALID_LOCATION "Invalid location."))
    (25 (values :INVALID_FIELDID "Invalid field."))
    (30 (values :INVALID_FRAMEID "Invalid jframeID."))
    (31 (values :NO_MORE_FRAMES "There are no more Java or JNI frames on the call stack."))
    (32 (values :OPAQUE_FRAME "Information about the frame is not available."))
    (33 (values :NOT_CURRENT_FRAME "Operation can only be performed on current frame."))
    (34 (values :TYPE_MISMATCH "The variable is not an appropriate type for the function used."))
    (35 (values :INVALID_SLOT "Invalid slot."))
    (40 (values :DUPLICATE "Item already set."))
    (41 (values :NOT_FOUND "Desired element not found."))
    (50 (values :INVALID_MONITOR "Invalid monitor."))
    (51 (values :NOT_MONITOR_OWNER "This thread doesn't own the monitor."))
    (52 (values :INTERRUPT "The call has been interrupted before completion."))
    (60 (values :INVALID_CLASS_FORMAT "The virtual machine attempted to read a class file and determined that the file is malformed or otherwise cannot be interpreted as a class file."))
    (61 (values :CIRCULAR_CLASS_DEFINITION "A circularity has been detected while initializing a class."))
    (62 (values :FAILS_VERIFICATION "The verifier detected that a class file, though well formed, contained some sort of internal invaluesistency or security problem."))
    (63 (values :ADD_METHOD_NOT_IMPLEMENTED "Adding methods has not been implemented."))
    (64 (values :SCHEMA_CHANGE_NOT_IMPLEMENTED "Schema change has not been implemented."))
    (65 (values :INVALID_TYPESTATE "The state of the thread has been modified, and is now invaluesistent."))
    (66 (values :HIERARCHY_CHANGE_NOT_IMPLEMENTED "A direct superclass is different for the new class version, or the set of directly implemented interfaces is different and canUnrestrictedlyRedefineClasses is false."))
    (67 (values :DELETE_METHOD_NOT_IMPLEMENTED "The new class version does not declare a method declared in the old class version and canUnrestrictedlyRedefineClasses is false."))
    (68 (values :UNSUPPORTED_VERSION "A class file has a version number not supported by this VM."))
    (69 (values :NAMES_DONT_MATCH "The class name defined in the new class file is different from the name in the old class object."))
    (70 (values :CLASS_MODIFIERS_CHANGE_NOT_IMPLEMENTED "The new class version has different modifiers and and canUnrestrictedlyRedefineClasses is false."))
    (71 (values :METHOD_MODIFIERS_CHANGE_NOT_IMPLEMENTED "A method in the new class version has different modifiers than its counterpart in the old class version and and canUnrestrictedlyRedefineClasses is false."))
    (99 (values :NOT_IMPLEMENTED "The functionality is not implemented in this virtual machine."))
    (100 (values :NULL_POINTER "Invalid pointer."))
    (101 (values :ABSENT_INFORMATION "Desired information is not available."))
    (102 (values :INVALID_EVENT_TYPE "The specified event type id is not recognized."))
    (103 (values :ILLEGAL_ARGUMENT "Illegal argument."))
    (110 (values :OUT_OF_MEMORY "The function needed to allocate memory and no more memory was available for allocation."))
    (111 (values :ACCESS_DENIED "Debugging has not been enabled in this virtual machine. JVMDI cannot be used."))
    (112 (values :VM_DEAD "The virtual machine is not running."))
    (113 (values :INTERNAL "An unexpected internal error has occurred."))
    (115 (values :UNATTACHED_THREAD "The thread being used to call this function is not attached to the virtual machine. Calls must be made from attached threads."))
    (500 (values :INVALID_TAG "Invalid object type id or class tag."))
    (502 (values :ALREADY_INVOKING "Previous invoke not complete."))
    (503 (values :INVALID_INDEX "Index is invalid."))
    (504 (values :INVALID_LENGTH "The length is invalid."))
    (506 (values :INVALID_STRING "The string is invalid."))
    (507 (values :INVALID_CLASS_LOADER "The class loader is invalid."))
    (508 (values :INVALID_ARRAY "The array is invalid."))
    (509 (values :TRANSPORT_LOAD "Unable to load the transport."))
    (510 (values :TRANSPORT_INIT "Unable to initialize the transport."))
    (511 (values :NATIVE_METHOD ""))
    (512 (values :INVALID_COUNT "The count is invalid."))
    (t (values nil "Unknown error."))))

(defconstant +single-step+	1)
(defconstant +breakpoint+	2)
(defconstant +frame-pop+	3)
(defconstant +exception+	4)
(defconstant +user-defined+	5)
(defconstant +thread-start+	6)
(defconstant +thread-death+	7)	;THREAD_END in the spec
(defconstant +class-prepare+	8)
(defconstant +class-unload+	9)
(defconstant +class-load+	10)
(defconstant +field-access+	20)
(defconstant +field-modification+ 21)
(defconstant +exception-catch+	30)
(defconstant +method-entry+	40)
(defconstant +method-exit+	41)
(defconstant +vm-start+		90)	;VM_INIT in the spec
(defconstant +vm-death+		99)
