;;;; Copyright (c) 2004 David Lichteblau <david@lichteblau.com>
;;;;
;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :jdii)

(define-class jdwp-server (jdwp-connection)
    (vm))

(defmethod initialize-instance :after ((instance jdwp-server) &key)
  (multiple-value-bind (a b c d e)
      (jdi:id-sizes (vm instance))
    (setf (field-id-size instance) a)
    (setf (method-id-size instance) b)
    (setf (object-id-size instance) c)
    (setf (type-id-size instance) d)
    (setf (frame-id-size instance) e)))

#+sbcl
(defun jdwp-listen (vm port)
  (let ((listenfd (make-instance 'sb-bsd-sockets:inet-socket
		    :type :stream
		    :protocol (sb-bsd-sockets:get-protocol-by-name "tcp"))))
    (setf (sb-bsd-sockets:sockopt-reuse-address listenfd) t)
    (unwind-protect
	(progn
	  (sb-bsd-sockets:socket-bind listenfd #(127 0 0 1) port)
	  (sb-bsd-sockets:socket-listen listenfd 5)
	  (loop
	    (let* ((sockfd (sb-bsd-sockets:socket-accept listenfd)))
	      (with-open-stream
		  (s (sb-bsd-sockets:socket-make-stream
		      sockfd
		      :input t
		      :output t
		      :element-type '(unsigned-byte 8)))
		(if (shake-hands-with-client s)
		    (return
		      (jdwp-loop (make-instance 'jdwp-server
				   :vm vm
				   :connection-stream s)))
		    (warn "Handshake error"))))))
      (sb-bsd-sockets:socket-close listenfd))))

(defun shake-hands-with-client (s)
  (let ((x (make-bytes (length *handshake*))))
    (read-sequence x s)
    (cond
      ((equalp x *handshake*)
	(write-sequence x s)
	(force-output s)
	t)
      (t
	nil))))

(defparameter *commands* (make-hash-table :test 'equal))

(defvar *trace-jdwp* nil)

(defun jdwp-loop (connection)
  (catch 'jdwp-loop
    (loop
      (fetch-message connection)
      (let* ((id (wp-read-int connection))
	     (flags (wp-read-byte connection))
	     (set (wp-read-byte connection))
	     (command (wp-read-byte connection))
	     (handler (gethash (cons set command) *commands*)))
	(declare (ignore flags))
	(unless handler
	  (error "JDWP command (~D ~D) not implemented" set command))
	(when *trace-jdwp*
	  (format t "~&;;; ~A (~D ~D)~%" handler set command))
	(funcall handler connection id)))))

(defmacro defcommand ((name set command) (connection id) &body body)
  `(SETF (GETHASH (CONS ,set ,command) *COMMANDS*)
	 (DEFUN ,name (,connection ,id)
	   (DECLARE (IGNORABLE ,id))
	   (RESET-OUT-BUFFER ,connection)
	   (LOCALLY
	     ,@body))))

(defun finish-reply (c id &optional error)
  (let ((buf (out-buffer c))
	(s (connection-stream c)))
    (let ((length (fill-pointer buf)))
      (setf (fill-pointer buf) 0)
      (wp-write-int length c)
      (wp-write-int id c)
      (wp-write-byte #x80 c)
      (wp-write-short (or error 0) c)
      (setf (fill-pointer buf) length))
    (write-sequence buf s)
    (force-output s)))



;;;; Virtual Machine Command Set

(defun class-tag (class)
  (etypecase class
    (jdi:array-type +typetag-array+)
    (jdi:interface-type +typetag-interface+)
    (jdi:class-type +typetag-class+)))

(defcommand (vm-all-classes 1 3) (s id)
  ;; lock
  (let ((classes (jdi:all-classes (vm s))))
    (wp-write-int (length classes) s)
    (dolist (class classes)
      (wp-write-byte (class-tag class) s)
      (wp-write-type (object->id class s) s)
      (wp-write-string (jdi:signature class) s)
      (wp-write-int +classstatus-initialized+ s) ;XXX
      )
    (finish-reply s id)))

(defcommand (vm-all-threads 1 4) (s id)
  (wp-write-int 1 s)			;one thread...
  (wp-write-int 42 s)			;random id
  (finish-reply s id))

(defcommand (vm-dispose 1 6) (s id)
  (jdi:dispose (vm s))
  (throw 'jdwp-loop nil))

(defcommand (vm-id-sizes 1 7) (s id)
  (multiple-value-bind (a b c d e)
      (jdi:id-sizes (vm s))
    (wp-write-int a s)
    (wp-write-int b s)
    (wp-write-int c s)
    (wp-write-int d s)
    (wp-write-int e s))
  (finish-reply s id))

(defcommand (vm-class-paths 1 13) (s id)
  (wp-write-string (jdi:base-directory (vm s)) s)
  (flet ((write-strings (list)
	   (wp-write-int (length list) s)
	   (dolist (d list)
	     (wp-write-string d s))))
    (write-strings (jdi:class-path (vm s)))
    (write-strings (jdi:bootstrap-class-path (vm s))))
  (finish-reply s id))



;;;; Reference Type Command Set

(defcommand (rt-signature 2 1) (s id)
  (let* ((class (id->object (wp-read-type s) s)))
    (wp-write-string (jdi:signature class) s)
    (finish-reply s id)))

(defcommand (rt-fields 2 4) (s id)
  (let* ((class (id->object (wp-read-type s) s))
	 (fields (jdi:fields class)))
    (wp-write-int (length fields) s)
    (dolist (field fields)
      (wp-write-field (object->id field s) s)
      (wp-write-string (jdi:name field) s)
      (wp-write-string (jdi:signature field) s)
      (wp-write-int (jdi:modifiers field) s))
    (finish-reply s id)))

(defcommand (rt-methods 2 5) (s id)
  (let* ((class (id->object (wp-read-type s) s))
	 (methods (jdi:methods class)))
    (wp-write-int (length methods) s)
    (dolist (method methods)
      (wp-write-method (object->id method s) s)
      (wp-write-string (jdi:name method) s)
      (wp-write-string (jdi:signature method) s)
      (wp-write-int (jdi:modifiers method) s))
    (finish-reply s id)))

(defcommand (rt-interfaces 2 10) (s id)
  (let* ((class (id->object (wp-read-type s) s))
	 (interfaces (jdi:interfaces class)))
    (wp-write-int (length interfaces) s)
    (dolist (i interfaces)
      (wp-write-type (object->id i s) s))
    (finish-reply s id)))



;;;; Class Type Command Set

(defcommand (ct-superclass 3 1) (s id)
  (let* ((class (id->object (wp-read-type s) s))
	 (superclass (jdi:superclass class)))
    (wp-write-type (object->id superclass s) s)
    (finish-reply s id)))



;;;; Event Request Command Set

(defvar *event-id* 0)

(defcommand (er-set 15 1) (s id)
  (let ((event-kind (wp-read-byte s))
	(suspend-policy (wp-read-byte s))
	(nmodifiers (wp-read-int s))
	(modifiers '()))
    (declare (ignore event-kind suspend-policy))
    (dotimes (x nmodifiers)
      (push (ecase (wp-read-byte s)
	      (1 (list :count (wp-read-int s)))
	      (2 (list :conditional (wp-read-int s)))
	      (3 (list :thread (wp-read-object s)))
	      (4 (list :class-only (wp-read-type s)))
	      (5 (list :class-match (wp-read-string s)))
	      (6 (list :class-exclude (wp-read-string s)))
	      (7 (list :location-only (error "ni") '(wp-read-location s)))
	      (8 (list :exception-only (list (wp-read-type s)
					     (wp-read-boolean s)
					     (wp-read-boolean s))))
	      (9 (list :field-only (list (wp-read-type s) (wp-read-field s))))
	      (10 (list :step (list (wp-read-object s)
				    (wp-read-int s)
				    (wp-read-int s))))
	      (11 (list :instance-only (wp-read-object s))))
	    modifiers))
    (wp-write-int (incf *event-id*) s)
    (finish-reply s id)
    (print (cons :event *event-id*))))



;;;; Thread Reference Command Set

(defcommand (tr-name 11 1) (s id)
  (wp-write-string "dummy" s)
  (finish-reply s id))

(defcommand (tr-thread-group 11 5) (s id)
  (assert (eq (wp-read-int s) 42))
  (wp-write-int 23 s)			;XXX
  (finish-reply s id))



;;;; Thread Group Reference Command Set

(defcommand (tg-name 12 1) (s id)
  (assert (eq (wp-read-int s) 23))
  (wp-write-string "dummy" s)
  (finish-reply s id))

(defcommand (tg-parent 12 2) (s id)
  (assert (eq (wp-read-int s) 23))
  (wp-write-int 0 s)
  (finish-reply s id))
