;;;; Copyright (c) 2004 David Lichteblau <david@lichteblau.com>
;;;;
;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :decaf)

(defmethod display ((object package-page) (frame decaf) *standard-output*)
  (terpri)
  (formatting-header (t :grey)
    (with-text-style (t (make-text-style :sans-serif :bold :large))
      (cond
	((name object)
	  (write-string "Package ")
	  (write-string (name object)))
	(t
	  (write-string "Package Hierarchy Root")))))
  (terpri)
  (let* ((prefix (concatenate 'string (name object) "."))
	 (classes (jdi:all-classes (virtual-machine frame)))
	 (classes
	  (remove-if-not (lambda (class)
			   (if (name object)
			       (jdii::starts-with-p (jdi:name class) prefix)
			       (null (position #\. (jdi:name class)))))
			 classes))
	 (classes (sort classes #'string-lessp :key #'jdi:name)))
    (with-indentation ()
      (dolist (class classes)
	(present class 'class-type)
	(terpri)))))

(defmethod display ((object (eql :packages)) (frame decaf) *standard-output*)
  (terpri)
  (formatting-header (t :grey)
    (with-text-style (t (make-text-style :sans-serif :bold :large))
      (write-string "Packages")))
  (terpri)
  (let* ((all-types (jdi:all-classes (virtual-machine frame)))
	 (classes (remove-if-not (rcurry #'typep 'jdi:class-type) all-types))
	 (names (mapcar (lambda (class)
			  (nth-value 1 (split-type-name (jdi:name class))))
			classes))
	 (names (remove-duplicates names :test #'equal))
	 (names (sort names #'string-lessp :key (lambda (x) (or x "")))))
    (with-indentation ()
      (dolist (name names)
	(with-output-as-presentation (t name 'package)
	  (if (zerop (length name))
	      (with-text-style
		  (t (make-text-style :sans-serif :italic :normal))
		(write-string "(Package Hierarchy Root)"))
	      (write-string name)))
	(terpri)))))

(defun format-member-header (str members icon &optional (underlinep t))
  (formatting-row (t)
    (formatting-cell ()
      (%draw-icon t (find-icon icon)))
    (formatting-cell ()
      (flet ((doit ()
	       (with-text-style (t (make-text-style :sans-serif :bold :normal))
		 (format t "~&~A~%~%" str))))
	(if underlinep
	    (surrounding-output-with-border (t :shape :underline)
	      (doit))
	    (doit)))))
  (unless members
    (formatting-row (t)
      (formatting-cell ())
      (formatting-cell ())
      (formatting-cell ()
	(with-text-style (t (make-text-style :sans-serif :italic :normal))
	  (format t "~&(none)~%"))))))

(defun present-class (thing &key short)
  (let* ((vm (virtual-machine *application-frame*))
	 (class
	  (if (stringp thing)
	      (car (jdi:classes-by-signature vm thing))
	      thing))
	 (jni-name (if class (jdi:signature class) thing))
	 (java-name (jdii::jni-to-java jni-name))
	 (str (if short (split-type-name java-name) java-name)))
    (cond
      ((or (eql (length jni-name) 1)
	   #+(or) (char= (char jni-name 0) #\[))
	(write-string str))
      (class
	(with-output-as-presentation (t class 'class-type)
	  (write-string str)))
      (t
	(with-drawing-options (t :ink +dark-grey+)
	  (write-string str))))))

(defun format-field-rows (frame object str test valuesp icon
			  &optional (underlinep t))
  (declare (ignore frame))
  (let* ((class (if (typep object 'jdi:object)
		    (jdi:reference-type object)
		    object))
	 (all-fields (if (show-inherited-p *application-frame*)
			 (jdi:all-fields class)
			 (jdi:fields class)))
	 (fields (sort (remove-if-not test all-fields)
		       #'string-lessp
		       :key #'jdi:name))
	 (values (if valuesp (jdi:field-values fields object) nil)))
    (format-member-header str fields icon underlinep)
    (jdii::for* ((field :in fields)
		 (value = (pop values))
		 (inheritedp = (eq class (jdi:declaring-type field)))
		 (face = (if inheritedp :bold nil)))
      (formatting-row (t)
	(formatting-cell (t)
	  #+(or)
	  (when (jdi:staticp field)
	    (with-text-style (t (make-text-style :sans-serif nil :small))
	      (write-string "static"))))
	(formatting-cell (t :align-x :right)
	  (with-text-style (t (make-text-style :sans-serif nil :small))
	    (present-class (jdi:signature field) :short t)))
	(with-output-as-presentation (t field 'jdi:field)
	  (formatting-cell (t)
	    (with-text-style (t (make-text-style :sans-serif face :normal))
	      (write-string (jdi:name field)))))
	(when valuesp
	  (formatting-cell (t)
	    (write-char #\=))
	  (formatting-cell (t)
	    (present value)))))))

(defun present-method (method &key class (face :bold))
  (with-text-style (t (make-text-style :sans-serif face :normal))
    (when class
      (when (eq class t)
	(setf class (jdi:declaring-type method)))
      (present-class class)
      (write-char #\.)))
  (with-text-style (t (make-text-style :sans-serif face :normal))
    (with-output-as-presentation (t method 'jdi:method)
      (if class
	  (surrounding-output-with-border
	      (t :shape :underline :move-cursor nil)
	    (write-string (jdi:name method)))
	  (write-string (jdi:name method))))
    (write-char #\())
  (let ((arguments (jdi:argument-type-names method)))
    (when arguments
      (present-class (car arguments) :short t)
      (dolist (argument (cdr arguments))
	(with-text-style (t (make-text-style :sans-serif face :normal))
	  (write-string ", "))
	(present-class argument :short t))))
  (with-text-style (t (make-text-style :sans-serif face :normal))
    (write-char #\))))

(defun format-method-rows (frame class str test)
  (declare (ignore frame))
  (let* ((all-methods (if (show-inherited-p *application-frame*)
			  (jdi:all-methods class :exclude-overriden t)
			  (jdi:methods class)))
	 (methods (sort (remove-if-not test all-methods)
			#'string-lessp
			:key #'jdi:name)))
    (format-member-header str methods :dot)
    (dolist (method methods)
      (formatting-row (t)
	(formatting-cell (t)
	  #+(or)
	  (when (jdi:staticp method)
	    (with-text-style (t (make-text-style :sans-serif nil :small))
	      (write-string "static"))))
	(formatting-cell (t :align-x :right)
	  (with-text-style (t (make-text-style :sans-serif nil :small))
	    (present-class (jdi:return-type-name method) :short t)))
	(formatting-cell (t)
	  (let ((inheritedp (eq class (jdi:declaring-type method))))
	    (present-method method :face (if inheritedp :bold nil))))))))

(defun format-class-header (frame class)
  (declare (ignore frame))
  (formatting-header (t :class)
    (multiple-value-bind (rname package)
	(split-type-name (jdi:name class))
      (when package
	(with-output-as-presentation (t package 'package)
	  (write-line package)))
      (with-output-as-presentation (t class 'class-type)
	(with-text-style (t (make-text-style :sans-serif :bold :large))
	  (etypecase class
	    (jdi:interface-type (write-string "Interface "))
	    (jdi:class-type (write-string "Class "))
	    (jdi:array-type (write-string "Array Class ")))
	  (write-line rname))))))

(defun format-metadata (frame class)
  (unless (typep class 'jdi:interface-type)
    (let ((superclass (jdi:superclass class)))
      (when superclass
	(with-text-style (t (make-text-style :sans-serif :italic nil))
	  (write-string "Superclass: "))
	(present-class superclass)
	(terpri))))
  (when (typep class 'jdi:array-type)
    (let ((element-type (or (jdi:component-type class)
			    ;; XXX
			    (jdi:component-signature class))))
      (with-text-style (t (make-text-style :sans-serif :italic nil))
	(write-string "Element type: "))
      (present-class element-type)
      (terpri)))
  (let* ((indentation "          ")	;...
	 (width (- (rectangle-width (window-viewport *standard-output*))
		   (* (text-size *standard-output* indentation) 2))))
    (flet ((present-class-list (header classes)
	     (when classes
	       (with-text-style (t (make-text-style :sans-serif :italic nil))
		 (write-string header))
	       (cond
		 ((cdr classes)
		   (terpri)
		   (filling-output (t :fill-width (list width :pixel)
				      :break-characters '(#\space)
				      :after-line-break indentation
				      :after-line-break-initially t)
		     ;; cannot use format-textual-list here because it doesn't
		     ;; seem to cooperate with the filling-output.
		     (dolist (class classes)
		       (unless (eq class (car classes))
			 (write-char #\,)
			 (write-char #\space))
		       (present-class class))))
		 (t
		   (present-class (car classes)))) 
	       (terpri))))
      (present-class-list (if (typep class 'jdi:class-type)
			      "Implemented interfaces: "
			      "Superinterfaces: ")
			  (jdi:interfaces class))
      (when (typep class 'jdi:class-type)
	(present-class-list "All currently loaded subclasses: "
			    (jdi:subclasses class)))
      (when (typep class 'jdi:interface-type)
	(present-class-list "All currently loaded subinterfaces: "
			    (jdi:subinterfaces class)))
      (when (typep class 'jdi:interface-type)
	(present-class-list "All currently loaded implementing classes: "
			    (remove-if-not
			     (lambda (c)
			       (and (typep c 'jdi:class-type)
				    (member class (jdi:interfaces c))))
			     (jdi:all-classes (virtual-machine frame)))))))
  (when (typep class 'jdi:class-type)
    (let ((pathname (java-source-file class)))
      (cond
	(pathname
	  (with-text-style (t (make-text-style :sans-serif :italic nil))
	    (write-string "Source file: "))
	  (present pathname))
	(t
	  (with-text-style (t (make-text-style :sans-serif :italic nil))
	    (write-string "Source file not in class path."))))
      (terpri))))

;;; .class-Files enthalten den relativen Namen zugehoerigen .java-Files,
;;; nicht aber den Pfad dazu.  Wir durchsuchen auf gut Glueck einfach den 
;;; Classpath, zuvor aber noch *SOURCE-CODE-PATH*.
;;;
;;; XXX wir muessten hier wenigstens den path.separator (sp?) abfragen.

(defvar *source-code-path* '()
  "List of directories to look in for source files.")

(defun compute-source-code-path (vm)
  (let ((base (concatenate 'string (jdi:base-directory vm) "/")))
    (with-collector ()
      (dolist (str (append *source-code-path*
			   (jdi:bootstrap-class-path vm)
			   (jdi:class-path vm)))
	(collect
	 (merge-pathnames
	  (cond
	    ((string-suffix-p str "/") str)
	    (t (concatenate 'string str "/")))
	  base))))))

(defun java-source-file (class)
  (let ((name0 (jdi:source-file class)))
    (if name0
	(let* ((package (nth-value 1 (split-type-name (jdi:name class))))
	       (name1 (if package
			  (concatenate 'string
			    (substitute #\/ #\. package)
			    "/"
			    name0)
			  name0))
	       (vm (jdi:virtual-machine class)))
	  (dolist (directory (compute-source-code-path vm))
	    (let ((name2 (merge-pathnames name1 directory)))
	      (when (probe-file name2)
		(return name2)))))
	nil)))

(defmethod display ((class jdi:reference-type) (frame decaf) *standard-output*)
  (terpri)
  (format-class-header frame class)
  (terpri)
  (format-metadata frame class)
  (terpri)
  (terpri)
  (formatting-table (t :x-spacing '(1 :character))
    (format-field-rows frame class "Static Fields" #'jdi:staticp t :dot)
    (write-empty-line)
    (format-field-rows frame class "Instance Fields" (complement #'jdi:staticp) nil :dot)
    (write-empty-line)
    (format-method-rows frame class "Static Methods" #'jdi:staticp)
    (write-empty-line)
    (format-method-rows frame class "Instance Methods" (complement #'jdi:staticp)))
  ;; make scroll bar behave: 
  (terpri))

(define-decaf-command (com-toggle-inherited-members :name t) ()
  (setf (show-inherited-p *application-frame*)
	(not (show-inherited-p *application-frame*))))
