;;;; Copyright (c) 2004 David Lichteblau <david@lichteblau.com>
;;;;
;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :decaf)

(define-presentation-type class-name () :inherit-from 'string)
(define-presentation-type class-type () :inherit-from 'class-name)

(define-presentation-type package-name () :inherit-from 'string)
(define-presentation-type package () :inherit-from 'package-name)

(define-presentation-type history-action ())
(define-presentation-type back-action ())
(define-presentation-type forward-action ())

(define-presentation-type line-of-code ())

(define-presentation-method present (object (type class-type) stream view &key)
  (declare (ignore view))
  (write-string (jdi:name object) stream))

(define-presentation-method present
    (object (type jdi:field) stream view &key)
  (declare (ignore view))
  (write-string (jdi:name object) stream))

(define-presentation-method present
    (object (type jdi:method) stream view &key)
  (declare (ignore view))
  (write-string (jdi:name object) stream))

(jdii::define-class package-page () (name))

(jdii::define-class file-page ()
  (name
   (from :initform nil)
   (to :initform nil)
   (highlight :initform '())))

(defclass event-handler (pane) ())

(define-application-frame decaf ()
    ((virtual-machine :initarg :virtual-machine :accessor virtual-machine)
     (history :initform (list :virtual-machine) :accessor history)
     (future :initform '() :accessor future)
     (show-inherited-p :initform t :accessor show-inherited-p)
     (resizedp :initform nil :accessor resizedp)
     (trace-requests :initform '() :accessor trace-requests)
     (fd-handler :initform nil :accessor fd-handler))
  (:pointer-documentation nil)		;depends on layout, see below
  (:menu-bar nil)			;depends on layout, see below
  (:panes
    (interactor :interactor
		:text-style (clim:make-text-style :sans-serif nil :normal))
    (main :application
	  :incremental-redisplay nil
	  :end-of-line-action :allow
	  :end-of-page-action :allow
	  :scroll-bars t
;;; 	     :display-time t
	  :text-style (clim:make-text-style :sans-serif nil :normal)
	  :display-function (lambda (frame stream)
			      (maybe-do-resize-hack frame)
			      (invoke-without-event-listener
			       (lambda ()
				 (display (car (history frame)) frame stream)))))
    (buttons :application
	     :display-function (lambda (frame stream)
				 (invoke-without-event-listener
				  #'display-button-bar
				  frame stream))
	     :scroll-bars nil
	     :background +light-grey+
	     :height 10)
    (vadjust clim-extensions:box-adjuster-gadget)
    #+(or) (hadjust clim-extensions:box-adjuster-gadget)
    (event-handler event-handler))
  (:layouts
   (default
    (vertically ()
      (clim-internals::make-menu-bar 'menubar-command-table)
      buttons
      main
      vadjust
      interactor
      (make-pane 'pointer-documentation-pane)))
    (child
     (vertically ()
       buttons
       main))
    (never
     event-handler)))

(defclass parent-decaf (decaf) ())

(defclass child-decaf (decaf) ())

(defmethod execute-frame-command ((frame child-decaf) command)
  (declare (ignore command))
  (let ((*application-frame* frame))
    (call-next-method)))

;;; Workaround: Without this method, user-specified geometry doesn't take
;;; effect until the mouse is moved over a pane.  Test case:
;;;   (decaf port :left 0 :top 0 :width 800 :height 600)
;;; Note that the same arguments work fine for the address book :-/
(defun maybe-do-resize-hack (frame)
  (unless (resizedp frame)
    (let* ((sheet (frame-top-level-sheet frame))
	   (region (sheet-region sheet)))
      (allocate-space sheet (rectangle-width region) (rectangle-height region)))
    (start-event-listener frame)
    (setf (resizedp frame) t)))

(make-command-table
 'decaf-command-table
 :errorp nil
 :menu '(#+(or) ("New Window" :command com-new-window)
	 ("Suspend Virtual Machine" :command com-suspend-virtual-machine)
	 ("Resume Virtual Machine" :command com-resume-virtual-machine)
	 ("" :command nil)		;:TYPE :DELIMITER would be nice
	 ("Quit" :command com-quit)))

(make-command-table
 'view-command-table
 :errorp nil
 :menu '(("Toogle Inherited Members" :command com-toggle-inherited-members)
	 ("" :command nil)		;:TYPE :DELIMITER would be nice
	 ("List Packages" :command com-list-packages)
	 ("Browse Thread Groups" :command com-browse-thread-groups)
	 ("List Threads" :command com-list-threads)))

(make-command-table 'menubar-command-table
		    :errorp nil
		    :menu '(("Decaf" :menu decaf-command-table)
			    ("View" :menu view-command-table)))

(defun display-button-bar (frame *standard-output*)
  (let* ((height (rectangle-height (sheet-region *standard-output*)))
	 (y (truncate (- height 17) 2)))
    (stream-increment-cursor-position *standard-output* 0 y)
    (flet ((draw-button (icon action)
	     (let ((x (stream-cursor-position *standard-output*)))
	       (draw-icon t (find-icon :bevel) :move nil)
	       (stream-increment-cursor-position *standard-output* 2 2)
	       (with-output-as-presentation (t nil action)
		 (draw-icon t (find-icon icon)))
	       (clim-lisp:setf (stream-cursor-position *standard-output*)
			       (values (+ x 17) y)))))
      (stream-increment-cursor-position *standard-output* 17 0)
      (draw-button :arrowdown 'history-action)
      (stream-increment-cursor-position *standard-output* 2 0)
      (draw-button :arrowleft 'back-action)
      (draw-button :arrowright 'forward-action)))
  (stream-increment-cursor-position *standard-output* 40 0)
  (with-text-style (t (make-text-style :sans-serif :bold nil))
    (let ((vm (virtual-machine frame)))
      (let* ((threads (jdi:all-threads vm))
	     (k (count-if #'jdi:suspendedp threads))
	     (n (length threads)))
	(when (plusp k)
	  (format t "Note: ~D/~D threads are currently suspended." k n))))))

(defmacro formatting-header ((stream icon-name) &body body)
  `(formatting-table (,stream :x-spacing '(1 :character))
     (formatting-row (,stream)
       (formatting-cell (,stream)
	 (%draw-icon ,stream (find-icon ,icon-name)))
       (formatting-cell (,stream :align-y :center)
	 ,@body))))

(defmacro with-indentation ((&optional (stream 't)) &body body)
  `(indenting-output (,stream '(2 :line))
     ,@body))

(defun split-type-name (str)
  (let ((n (position #\. str :from-end t)))
    (values (subseq str (if n (1+ n) 0))
	    (if n (subseq str 0 n) nil))))

(defvar *icons* '())

(defparameter *icon-directory* "/home/david/src/cloak/decaf/")

(defun find-icon (name)
  (let ((cons (assoc name *icons*)))
    (unless cons
      (let* ((pathname
	      (make-pathname
	       :name (string-downcase name)
	       :type "xpm"
	       :defaults *icon-directory*))
	     (xpm (climi::xpm-parse-file pathname)))
	(setf cons (cons name xpm))
	(push cons *icons*)))
    (cdr cons)))

(defun %draw-icon (stream icon)
  ;; XXX
  (draw-icon stream icon)
  (write-string " "))

(defun write-empty-line ()
  ;; force vertical space
  (formatting-row (t)
    (formatting-cell (t)
      (write-char #\space)
      (terpri))))

(defun draw-icon (stream pattern &key (extra-spacing 0) (move t))
  (let ((stream (if (eq stream T) *standard-output* stream)))
    (multiple-value-bind (x y)
        (stream-cursor-position stream)
      (draw-pattern* stream pattern x y)
      (when move
	(stream-increment-cursor-position stream (+ (pattern-width pattern) extra-spacing) 0)))))

#+(or)
(define-decaf-command (com-new-window :name t) ()
  (unless (fboundp 'climi::frame-children)
    (error "Sorry, that command requires patch multi-frame.diff."))
  (let* ((parent *application-frame*)
	 (child
	  (make-application-frame
	   'child-decaf
	   :virtual-machine (virtual-machine parent)
	   :parent-frame parent
	   :calling-frame parent
	   :current-layout 'child
	   :pointer-documentation nil)))
    (adopt-frame (frame-manager parent) child)
    (enable-frame child)))

(define-decaf-command (com-clear-listener :name t) ()
  (clim:window-clear *standard-input*))

(defmethod frame-standard-output ((frame decaf))
  (climi::find-pane-of-type (frame-panes frame) 'interactor-pane))



;;;; History

(defun go-to (object)
  (push object (history *application-frame*))
  (setf (future *application-frame*) nil))

(define-decaf-command (com-back :name t)
    ((foo 'back-action :gesture :select))
  foo
  (if (cdr (history *application-frame*))
      (push (pop (history *application-frame*))
	    (future *application-frame*))
      (write-string "At beginning of history.")))

(define-decaf-command (com-forward :name t)
    ((foo 'forward-action :gesture :select))
  foo
  (if (future *application-frame*)
      (push (pop (future *application-frame*))
	    (history *application-frame*))
      (write-string "At end of history.")))

(define-decaf-command (com-history :name t)
    ((foo 'history-action :gesture :select))
  foo
  (let ((frame *application-frame*))
    (destructuring-bind (&optional history future)
	(clim:menu-choose
	 (jdii::with-collector ()
	   (jdii::for* ((future = (future frame) :then (cons entry future))
			(history :on (history frame))
			(entry = (car history)))
	     (jdii::collect
	      (list (etypecase entry
		      (package-page
			(concatenate 'string "Package " (name entry)))
		      ((eql :packages)
			"Package List")
		      ((eql :thread-groups)
			"Top Level Thread Groups")
		      ((eql :threads)
			"All Threads")
		      ((eql :virtual-machine)
			"Virtual Machine Details")
		      (file-page
			(concatenate 'string "File " (namestring (name entry))))
		      (jdi:reference-type
			(concatenate 'string "Class " (jdi:name entry)))
		      (jdi:thread-group
			(concatenate 'string "Thread Group " (jdi:name entry)))
		      (jdi:thread
			(concatenate 'string "Thread " (jdi:name entry)))
		      (jdi:object
			entry))
		    :value (list history future))))))
      (when history
	(setf (history frame) history)
	(setf (future frame) future)))))

(define-decaf-command (com-select-class :name t)
    ((class 'class-name :gesture :select))
  (when (stringp class)
    (let ((name class)
	  (jni-name (format nil "L~A;" class)))
      (setf class (find-if (lambda (mirror)
			     (or (equal name (jdi:name mirror))
				 (equal jni-name (jdi:signature mirror))
				 (equal name (jdi:signature mirror))))
			   (jdi:all-classes
			    (virtual-machine *application-frame*))))
      (unless class
	(format t "No class named ~S is currently loaded." name)
	(return-from com-select-class))))
  (go-to class))

(define-decaf-command (com-select-package :name t)
    ((package 'package-name :gesture :select))
  (go-to (make-instance 'package-page :name package)))

(define-presentation-type thread-name () :inherit-from 'string)
(define-presentation-type thread () :inherit-from 'thread-name)

(define-decaf-command (com-select-thread :name t)
    ((thread 'thread-name :gesture :select))
  (when (stringp thread)
    (let ((name thread))
      (setf thread (find thread
			 (jdi:all-threads
			  (virtual-machine *application-frame*))
			 :key #'jdi:name
			 :test #'equal))
      (unless thread
	(format t "Thread ~S not found." name)
	(return-from com-select-thread))))
  (go-to thread))

(define-presentation-type thread-group-name () :inherit-from 'string)
(define-presentation-type thread-group () :inherit-from 'thread-group-name)

(define-decaf-command (com-select-thread-group :name "Select Thread-Group")
    ((group 'thread-group-name :gesture :select))
  (when (stringp group)
    (let ((name group))
      (setf group (labels ((recurse (group)
			     (terpri)
			     (if (equal name (jdi:name group))
				 group
				 (recurse-list (jdi:child-groups group))))
			   (recurse-list (list)
			     (find-if #'recurse list)))
		    (recurse-list (jdi:top-level-thread-groups
				   (virtual-machine *application-frame*)))))
      (unless group
	(format t "Thread-Group ~S not found." name)
	(return-from com-select-thread-group))))
  (go-to group))

(define-decaf-command (com-select-object :name t)
    ((object 'jdi:object :gesture :select))
  (go-to object))

(define-decaf-command (com-select-file :name t)
    ((pathname 'pathname :gesture :select))
  (go-to (make-instance 'file-page :name (pathname pathname))))

(define-decaf-command (com-select-region :name t)
    ((file 'file-page :gesture :select))
  (go-to file))

(define-decaf-command (com-open-in-emacs :name t)
    ((line 'line-of-code :gesture :select))
  (destructuring-bind (file lineno) line
    (cond
      ((null (find-package :lep))
	(write-line "Sorry, ELI not loaded."))
      ((not (funcall (find-symbol "LEP-IS-RUNNING" :lep)))
	(write-line "Not connected to Emacs."))
      (t
	(funcall (find-symbol "FIND-FILE-IN-EDITOR" :lep)
		 :pathname (name file))
	(funcall (find-symbol "EVAL-IN-EMACS" :lep)
		 (format nil "(goto-line ~D)" lineno))))))

(define-decaf-command (com-select-method :name t)
    ((method 'jdi:method :gesture :select))
  (let ((pathname (java-source-file (jdi:declaring-type method)))
	(numbers (mapcar #'jdi:line-number (jdi:lines method))))
    (unless pathname
      (with-text-style (t (make-text-style :sans-serif :italic nil))
	(write-line "Sorry, source file not in class path."))
      (return-from com-select-method))
    (go-to (make-instance 'file-page
	     :name (pathname pathname)
	     :highlight numbers
	     ;; XXX wir zeigen immer ein bisschen context an.  Um den Header
	     ;; einen Methode vollstaendig anzeigen zu koennen, brauchen wir
	     ;; aber etwas mehr Intelligenz.  Man koennte z.B. nach einer
	     ;; geschweiften Klammer suchen.
	     :from (1- (reduce #'min numbers))
	     :to (+ (reduce #'max numbers) 2)))))

(define-decaf-command (com-list-packages :name t) ()
  (go-to :packages))

(define-decaf-command (com-list-threads :name t) ()
  (go-to :threads))

(define-decaf-command (com-browse-thread-groups :name t) ()
  (go-to :thread-groups))

(defun print-status ()
  (let ((vm (virtual-machine *application-frame*)))
    (let* ((threads (jdi:all-threads vm))
	   (k (count-if #'jdi:suspendedp threads))
	   (n (length threads)))
      (with-text-style
	  (t (make-text-style :sans-serif (if (plusp k) :bold nil) nil))
	(format t "~&Status: ~D/~D threads are currently suspended.~%" k n)))))

(define-decaf-command (com-suspend-virtual-machine :name t) ()
  (jdi:suspend (virtual-machine *application-frame*))
  (print-status))

(define-decaf-command (com-resume-virtual-machine :name t) ()
  (jdi:resume (virtual-machine *application-frame*))
  (print-status))

(define-decaf-command (com-quit :name t) ()
  (frame-exit *application-frame*))

(define-decaf-command (com-test :name t) ()
  (prin1 (accept 'class-type)))

(define-decaf-command (com-noop :name t) ()
  nil)

#+(or)
(define-decaf-command (com-test2 :name t) ()
  (print
   (accepting-values ()
     (accept 'class-type))))



;;;; Virtual Machine

(define-decaf-command (com-describe-virtual-machine :name t) ()
  (go-to :virtual-machine))

(defmethod display
    ((object (eql :virtual-machine)) (frame decaf) *standard-output*)
  (let ((vm (virtual-machine frame)))
    (terpri)
    (with-text-style (t (make-text-style :sans-serif :bold :large))
      (write-line "Virtual Machine Details"))
    (terpri)
    (with-text-style (t (make-text-style :sans-serif :bold nil))
      (write-string (jdi:name vm))
      (write-string ", version ")
      (write-string (jdi:version vm))
      (terpri))
    (write-line (jdi:description vm))
    (terpri)
    (with-text-style (t (make-text-style :sans-serif :bold nil))
      (write-line "Class-Paths"))
    (write-line "Bootstrap Class-Path:")
    (with-indentation ()
      (unless (jdi:bootstrap-class-path vm)
	(write-line "(empty)"))
      (with-text-style (t (make-text-style :fixed nil nil))
	(mapc #'write-line (jdi:bootstrap-class-path vm))))
    (write-line "User Class-Path:")
    (with-indentation ()
      (unless (jdi:class-path vm)
	(write-line "(empty)"))
      (with-text-style (t (make-text-style :fixed nil nil))
	(mapc #'write-line (jdi:class-path vm))))
    (terpri)
    (with-text-style (t (make-text-style :sans-serif :bold nil))
      (write-line "JDWP"))
    (format t "Version: ~D.~D~%"
	    (car (jdi:jdwp-version vm))
	    (cdr (jdi:jdwp-version vm)))
    (format t "ID sizes: ~D (fields), ~D (methods), ~D (objects), ~D (types), ~D (frames)"
	    (jdii::field-id-size vm)
	    (jdii::method-id-size vm)
	    (jdii::object-id-size vm)
	    (jdii::type-id-size vm)
	    (jdii::frame-id-size vm))
    (fresh-line)
    (terpri)
    (with-text-style (t (make-text-style :sans-serif :bold nil))
      (write-line "Capabilities"))
    (formatting-table ()
      (map nil (lambda (name)
		 (when name
		   (formatting-row ()
		     (formatting-cell (t :align-y :center)
		       (let* ((value (jdi:get-capability vm name))
			      (icon (if value :checkboxdown :checkbox)))
			 (draw-icon t (find-icon icon))))
		     (formatting-cell (t :align-y :center)
		       (format t "~:(~A~)"
			       (substitute #\space #\- (symbol-name name)))))))
	   jdii::+capability-names+))
    (fresh-line)))



;;;; startup

(defun request-event (callback class &rest initargs)
  (jdi:request-event (virtual-machine *application-frame*)
		     (apply #'make-instance
			    class
			    :plist (list :callback callback)
			    initargs)))

(defun trace-print (frame)
  (let* ((location (jdi:location frame))
	 (method (jdi:method location)))
    (with-text-style (t (make-text-style :sans-serif :bold :normal))
      (present-class (jdi:type location))
      (write-char #\.))
    (with-text-style (t (make-text-style :sans-serif :bold :normal))
      (with-output-as-presentation (t method 'jdi:method)
	(write-string (jdi:name method)))
      (write-char #\())
    (multiple-value-bind (locals values) (fetch-local-variables frame)
      (declare (ignore locals))
      (cond
	(values
	  (present (car values))
	  (dolist (argument (cdr values))
	    (with-text-style (t (make-text-style :sans-serif :bold :normal))
	      (write-string ", "))
	    (present argument)))
	(t
	  (with-text-style (t (make-text-style :sans-serif :italic nil))
	    (write-string "local variables not available")))))
    (with-text-style (t (make-text-style :sans-serif :bold :normal))
      (write-char #\)))))

(defun trace-callback (event)
  (fresh-line)
  (write-string "Entered method: ")
  (jdi:jdwp-handler-case
      (trace-print (car (jdi:frames (jdi:thread event))))
    (:THREAD_NOT_SUSPENDED ()
      (warn "Oops, thread has already been suspended.")))
  (fresh-line))

(define-decaf-command (com-trace-method :name t)
    ((method 'jdi:method :gesture :right))
  (push (request-event 'trace-callback
		       'jdi:breakpoint-request
		       :suspend-policy :thread
		       :location (jdi:location method))
	(trace-requests *application-frame*)))

(define-decaf-command (com-untrace-all :name t) ()
  (let ((vm (virtual-machine *application-frame*)))
    (dolist (request (trace-requests *application-frame*))
      (jdi:cancel-request vm request))))

(defun breakpoint-callback (event)
  (fresh-line)
  (setf (getf (jdi:plist (jdi:request event)) :keep-suspended) t)
  (let* ((thread (jdi:thread event))
	 (frame (car (jdi:frames thread))))
    (with-text-style (t (make-text-style :sans-serif :bold nil))
      (write-string "Breakpoint reached in "))
    (with-text-style (t (make-text-style :fixed nil nil))
      (with-output-as-presentation (t thread 'thread)
	(format t "thread ~A" (jdi:name thread))))
    (with-text-style (t (make-text-style :sans-serif :bold nil))
      (write-string ", method "))
    (let ((method (jdi:method (jdi:location frame))))
      (with-text-style (t (make-text-style :fixed nil nil))
	(with-output-as-presentation (t method 'jdi:method)
	  (write-string (jdi:name method)))))
    (fresh-line)
    (with-indentation ()
      (print-stack-frame frame :suppress-warning t))))

(define-decaf-command (com-set-breakpoint-on-method :name t)
    ((method 'jdi:method :gesture :right))
  (push (request-event 'breakpoint-callback
		       'jdi:breakpoint-request
		       :suspend-policy :thread
		       :location (jdi:location method))
	(trace-requests *application-frame*)))

(defun process-events (frame)
  (for ((set = (jdi:read-event-set (virtual-machine frame)))
	:while set)
    (dolist (event (jdi:events set))
      (let* ((request (jdi:request event))
	     (callback (getf (jdi:plist request) :callback)))
	(if callback
	    (funcall callback event)
	    (warn "ignoring unsolicited event ~A" event))))
    (unless (some (lambda (event)
		    (getf (jdi:plist (jdi:request event)) :keep-suspended))
		  (jdi:events set))
      (jdi:resume set))))

(defmethod handle-event ((client event-handler) (event timer-event))
  (process-events *application-frame*)
  (start-event-listener *application-frame*))

(defun inject-event (frame)
  (let* ((event-handler (find-pane-named frame 'event-handler))
	 (event (make-instance 'timer-event :sheet event-handler)))
    (climi::event-queue-append (sheet-event-queue event-handler) event))
  (remove-fd-handler (fd-handler frame)))

(defun start-event-listener (frame)
  (let ((stream (jdii::connection-stream (virtual-machine frame))))
    (flet ((callback (&rest args)
             (declare (ignore args))
             (inject-event frame)))
      (setf (fd-handler frame) (add-fd-handler stream :input #'callback)))))

(defun invoke-without-event-listener (fn &rest args)
  (let ((frame *application-frame*))
    (if (fd-handler frame)
	(let ((p (fd-handler frame)))
	  (setf (fd-handler frame) nil)
	  (remove-fd-handler p)
	  (unwind-protect
	      (apply fn args)
	    (start-event-listener frame)))
	(apply fn args))))

(defun wrap-invoke-without-event-listener (fn)
  (lambda (&rest args) (apply #'invoke-without-event-listener fn args)))

(defmethod execute-frame-command :around ((frame decaf) command)
  (declare (ignore command))
  (invoke-without-event-listener #'call-next-method))

(defun decaf (&optional (port 8000) &rest initargs &key left top right bottom)
  (mapc #'destroy-port climi::*all-ports*)
  (setf climi::*all-ports* nil)
  (jdi:with-virtual-machine (vm (jdi:jdwp-connect "localhost" port))
    (let ((frame (apply #'make-application-frame
			#+(or) 'parent-decaf
			'decaf
			:virtual-machine vm
			:left (or left 0)
			:top (or top 0)
			:right (or right 1014)
			:bottom (or bottom 758)
			initargs)))
      (unwind-protect
	  (run-frame-top-level frame)
	(when (fd-handler frame)
	  (remove-fd-handler (fd-handler frame)))))))
