;;;; Copyright (c) 2004 David Lichteblau <david@lichteblau.com>
;;;;
;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :decaf)

(defmethod display ((object (eql :threads)) (frame decaf) *standard-output*)
  (terpri)
  (formatting-header (t :grey)
    (with-text-style (t (make-text-style :sans-serif :bold :large))
      (write-string "All Threads")))
  (terpri)
  (formatting-table (t :x-spacing '(1 :character))
    (with-text-style (t (make-text-style :sans-serif :bold nil))
      (formatting-row (t)
	(formatting-cell (t))
	(formatting-cell (t :min-width '(10 :character))
	  (write-string "Thread"))
	(formatting-cell (t :min-width '(7 :character))
	  (write-string "Group"))
	(formatting-cell (t :min-width '(7 :character))
	  (write-string "Status"))
	(formatting-cell (t)
	  (write-string "Suspended?"))))
    (dolist (thread (sort (jdi:all-threads (virtual-machine frame))
			  #'string-lessp
			  :key #'jdi:name))
      (formatting-row (t)
	(formatting-cell (t :align-y :center)
	  (draw-icon t (find-icon :red-dot)))
	(formatting-cell (t :align-y :center)
	  (let ((name (jdi:name thread)))
	    (with-output-as-presentation (t thread 'thread)
	      (write-string name))))
	(formatting-cell (t :align-y :center)
	  (let* ((group (jdi:group thread))
		 (name (jdi:name group)))
	    (with-output-as-presentation (t group 'thread-group)
	      (write-string name))))
	(formatting-cell (t :align-y :center)
	  (with-text-style (t (make-text-style :fixed nil nil))
	    (princ (jdi:status thread))))
	(formatting-cell (t :align-y :center)
	  (when (jdi:suspendedp thread)
	    (with-text-style (t (make-text-style :sans-serif '(:bold :italic) nil))
	      (write-string "Yes"))))))))

(defmethod display
    ((object (eql :thread-groups)) (frame decaf) *standard-output*)
  (terpri)
  (formatting-header (t :grey)
    (with-text-style (t (make-text-style :sans-serif :bold :large))
      (write-string "Top Level Thread Groups")))
  (terpri)
  (with-indentation ()
    (dolist (group (jdi:top-level-thread-groups (virtual-machine frame)))
      (let ((name (jdi:name group)))
	(with-output-as-presentation (t group 'thread-group)
	  (write-string name)))
      (terpri))))

(defmethod display
    ((group jdi:thread-group) (frame decaf) *standard-output*)
  (terpri)
  (formatting-header (t :green)
    (with-text-style (t (make-text-style :sans-serif :bold :large))
      (with-output-as-presentation (t group 'thread-group)
	(write-string "Thread Group ")
	(write-string (jdi:name group)))))
  (terpri)
  (with-text-style (t (make-text-style :sans-serif :italic nil))
    (write-string "Parent Group: "))
  (let ((parent (jdi:parent group)))
    (if parent
	(with-output-as-presentation (t parent 'thread-group)
	  (write-string (jdi:name parent)))
	(with-text-style (t (make-text-style :sans-serif :italic :normal))
	  (format t "(none)~%"))))
  (fresh-line)
  (with-text-style (t (make-text-style :sans-serif :italic nil))
    (write-string "Instance of: "))
  (present-class (jdi:reference-type group))
  (terpri)
  (terpri)
  (flet ((header (title icon)
	   (formatting-table (t)
	     (formatting-row (t)
	       (formatting-cell (t)
		 (%draw-icon t (find-icon icon)))
	       (formatting-cell (t)
		 (with-text-style
		     (t (make-text-style :sans-serif :bold :normal))
		   (format t "~&~A~%~%" title)))))
	   (fresh-line)))
    (let ((groups (jdi:child-groups group)))
      (header "Child Groups" :red-dot)
      (with-indentation ()
	(if groups
	    (dolist (group groups)
	      (let ((name (jdi:name group)))
		(with-output-as-presentation (t group 'thread-group)
		  (write-string name)))
	      (terpri))
	    (with-text-style (t (make-text-style :sans-serif :italic :normal))
	      (format t "~&(none)~%")))))
    (terpri)
    (let ((threads (jdi:child-threads group)))
      (header "Child Threads" :red-dot)
      (with-indentation ()
	(if threads
	    (dolist (thread threads)
	      (let ((name (jdi:name thread)))
		(with-output-as-presentation (t thread 'thread)
		  (write-string name)))
	      (terpri))
	    (with-text-style (t (make-text-style :sans-serif :italic :normal))
	      (format t "~&(none)~%"))))))
  (terpri)
  (surrounding-output-with-border (t :shape :underline)
    (with-text-style (t (make-text-style :sans-serif :bold nil))
      (write-string "Fields:")))
  (fresh-line)
  (display-nonarray group frame nil)
  (terpri))

(defun fetch-local-variables (frame)
  "Return as multiple values (a) the list local variables, (b) a list
   of variable values, and (c) an optional warning message.  Either (b) or
   both of (a) and (b) can be NIL if the requested information is not
   available."
  (let* ((location (jdi:location frame))
	 (locals (jdi:guess-local-variables
		  (jdi:method location)
		  (jdi:type location)))
	 (warning nil)
	 values)
    (with-text-style (t (make-text-style :sans-serif nil :small))
      (jdi:jdwp-handler-case
	  (setf locals (jdi:local-variables (jdi:method location)))
	(:NATIVE_METHOD ()
	  (setf warning "(Guessing local variable data for native method.)"))
	(:ABSENT_INFORMATION ()
	  (setf warning "(Guessing local variable data for class without debugging information.)"))))
    (setf values (jdi:jdwp-handler-case
		     (jdi:get-values frame locals)
		   (:OPAQUE_FRAME () nil)))
    (values locals values warning)))

(defun print-stack-frame (frame &key suppress-warning)
  (multiple-value-bind (locals values warning)
      (fetch-local-variables frame)
    (when (and warning (not suppress-warning))
      (write-line warning)
      (terpri))
    (dolist (local locals)
      (with-text-style (t (make-text-style :fixed nil nil))
	(present-class (jdi:signature local) :short t)
	(write-char #\space)
	(cond
	  ((jdi:name local)
	    (write-string (jdi:name local)))
	  (t
	    (write-string "<local")
	    (princ (jdii::id local))
	    (write-char #\>)))
	(write-string " = "))
      (if values
	  (with-text-style (t (make-text-style :fixed nil nil))
	    (present (pop values)))
	  (with-text-style (t (make-text-style :sans-serif :italic nil))
	    (write-string "(not available)")))
      (terpri))))

(defmethod display
    ((thread jdi:thread) (frame decaf) *standard-output*)
  (terpri)
  (formatting-header (t :red)
    (write-string "Thread")
    (terpri)
    (with-text-style (t (make-text-style :sans-serif :bold :large))
      (with-output-as-presentation (t thread 'thread)
	(write-string (jdi:name thread)))))
  (terpri)
  (with-text-style (t (make-text-style :sans-serif :italic nil))
    (write-string "Thread Group: "))
  (let ((group (jdi:group thread)))
    (with-output-as-presentation (t group 'thread-group)
      (write-string (jdi:name group))))
  (terpri)
  (with-text-style (t (make-text-style :sans-serif :italic nil))
    (write-string "Status: "))
  (princ (jdi:status thread))
  (terpri)
  (with-text-style (t (make-text-style :sans-serif :italic nil))
    (if (jdi:suspendedp thread)
	(write-line "This thread is currently suspended.")
	(write-line "Not suspended.")))
  (with-text-style (t (make-text-style :sans-serif :italic nil))
    (write-string "Instance of: "))
  (present-class (jdi:reference-type thread))
  (terpri)
  (terpri)
  (let ((climi::*pane-realizer* *DEFAULT-FRAME-MANAGER*)
	(suspendedp (jdi:suspendedp thread)))
    (assert climi::*pane-realizer*)
    (with-output-as-gadget (t)
      (make-pane 'push-button-pane
		 :label (if suspendedp
			    "Resume thread"
			    "Suspend thread")
		 :activate-callback
		 (wrap-invoke-without-event-listener
		  (lambda (button)
		    (declare (ignore button))
		    (if suspendedp
			(jdi:resume thread)
			(jdi:suspend thread))
		    (let ((pane (find-pane-named frame 'main)))
		      (window-clear pane)
		      (redisplay-frame-pane frame pane :force-p t)))))))
  (fresh-line)
  (terpri)
  (when (jdi:suspendedp thread)
    (surrounding-output-with-border (t :shape :underline)
      (with-text-style (t (make-text-style :sans-serif :bold :normal))
	(write-line "Stack")))
    (terpri)
    (formatting-table (t)
      (let ((frames (jdi:frames thread)))
	(unless frames
	  (write-string "Hmm, this thread has no frames!"))
	(dolist (frame frames)
	  (let ((location (jdi:location frame)))
	    (formatting-row (t)
	      (formatting-cell (t)
		(%draw-icon t (find-icon :dot)))
	      (formatting-cell (t)
		(present-method (jdi:method location) :class (jdi:type location))))
	    (formatting-row (t)
	      (formatting-cell (t))
	      (formatting-cell (t)
		(print-stack-frame frame)))
	    (formatting-row (t)
	      (formatting-cell (t)
		(write-char #\space))))))))
  (fresh-line)
  (surrounding-output-with-border (t :shape :underline)
    (with-text-style (t (make-text-style :sans-serif :bold nil))
      (write-string "Fields:")))
  (terpri)
  (display-nonarray thread frame nil)
  (terpri))

(define-decaf-command (com-suspend-thread :name t)
    ((thread 'thread :gesture :right))
  (jdi:suspend thread))

(define-decaf-command (com-resume-thread :name t)
    ((thread 'thread :gesture :right))
  (jdi:resume thread))
