;;;; Copyright (c) 2003,2004 David Lichteblau <david@lichteblau.com>
;;;;
;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(in-package :dlisp)

(defun nanp (float)
  (sb-ext:float-nan-p float))

(defun |java/lang/Double.doubleToRawLongBits(D)| (d)
  (canonicalize-long
   (logior (ash (sb-kernel:double-float-high-bits d) 32)
	   (sb-kernel:double-float-low-bits d))))
(setf (fdefinition '|java/lang/VMDouble.doubleToRawLongBits(D)|)
      #'|java/lang/Double.doubleToRawLongBits(D)|)

(defun |java/lang/Double.longBitsToDouble(J)| (l)
  (sb-kernel:make-double-float
   (canonicalize-int (ldb (byte 32 32) l))
   (canonicalize-n-unsigned (ldb (byte 32 0) l) 32)))
(setf (fdefinition '|java/lang/VMDouble.longBitsToDouble(J)|)
      #'|java/lang/Double.longBitsToDouble(J)|)

(defun |java/lang/Float.floatToRawIntBits(F)| (d)
  (canonicalize-int (sb-kernel:single-float-bits d)))
(setf (fdefinition '|java/lang/VMFloat.floatToRawIntBits(F)|)
      #'|java/lang/Float.floatToRawIntBits(F)|)

(defun |java/lang/Float.intBitsToFloat(I)| (l)
  (sb-kernel:make-single-float (canonicalize-int l)))
(setf (fdefinition '|java/lang/VMFloat.intBitsToFloat(I)|)
      #'|java/lang/Float.intBitsToFloat(I)|)

(defun read-sequence-partial-fill (seq stream start end)
  (%read-sequence-partial-fill seq stream start end))

(defun get-host-by-address (ipaddr-array)
  (handler-case
      (sb-bsd-sockets:host-ent-name
       (sb-bsd-sockets:get-host-by-address
	ipaddr-array))
    (sb-bsd-sockets:host-not-found-error ()
      (throw-exception "java/net/UnknownHostException"))))

(defun get-host-by-name (name)
  (handler-case
      (sb-bsd-sockets:host-ent-address (sb-bsd-sockets:get-host-by-name name))
    (sb-bsd-sockets:host-not-found-error ()
      (throw-exception "java/net/UnknownHostException"))))

(defun make-socket (inaddr-array port)
  (let* ((protocol (sb-bsd-sockets:get-protocol-by-name "tcp"))
	 (socket (make-instance 'sb-bsd-sockets:inet-socket
		   :type :stream
		   :protocol protocol)))
    (sb-bsd-sockets:socket-connect socket inaddr-array port)
    socket))

(defun socket-stream (socket &key (element-type '(signed-byte 8)))
  (sb-bsd-sockets:socket-make-stream
   socket
   :input t
   :output t
   :element-type element-type))

(defconstant +socket-error+ 'sb-bsd-sockets::socket-error)

(defun set-socket-option (socket option value)
  (ecase option
    (#.+plainsocketimpl-keep-alive+
      (setf (sb-bsd-sockets:sockopt-keep-alive socket) value))
    (#.+plainsocketimpl-nodelay+
      (setf (sb-bsd-sockets:sockopt-tcp-nodelay socket) value))))

(defun gc (&optional full)
  (sb-ext:gc :full full))

(defun string-to-utf-8 (string)
  (map 'vector #'char-code string))

(defun utf-8-to-string (bytes)
  (map 'string #'code-char bytes))

(defun decode-pseudo-utf-8 (lisp-string)
  (parse-utf-8 lisp-string))

(defun compile-labeled-function (name lambda-expression)
  (fake-labeled-function name lambda-expression))

(defvar *standard-floating-point-modes* nil)

(defun disable-floating-point-traps ()
  (unless *standard-floating-point-modes*
    (setf *standard-floating-point-modes* (sb-int:get-floating-point-modes)))
  (sb-int:set-floating-point-modes :traps nil))

(defun enable-floating-point-traps ()
  (assert *standard-floating-point-modes*)
  (apply #'sb-int:set-floating-point-modes *standard-floating-point-modes*))

(sb-alien:define-alien-routine "fmod" double-float
  (f double-float)
  (g double-float))

(sb-alien:define-alien-routine "fmodf" single-float
  (f single-float)
  (g single-float))

(defun drem (f g)
  (if (zerop g)
      +nan+
      (fmod f g)))

(defun frem (f g)
  (if (zerop g)
      (float +nan+ 1.0s0)
      (fmodf f g)))

(defun directoryp (pathname)
  (eq :directory (sb-unix:unix-file-kind pathname)))

(defun wait-for-input-available (stream)
  (when (streamp stream)
    (setf stream (sb-impl::fd-stream-fd stream)))
  (sb-sys:wait-until-fd-usable stream :input))

;; OK, so this is silly.  SERVE-EVENT doesn't work with CLX and/or McCLIM,
;; so we emulate it.  Or perhaps I need to find out why it doesn't work for
;; me, perhaps I'm misusing something.

(defun add-fd-handler (stream direction function)
  #+(or)
  (sb-sys:add-fd-handler (sb-impl::fd-stream-fd stream) direction function)
  (assert (eq direction :input))
  (sb-thread:make-thread
   (lambda ()
     (while (open-stream-p stream)
       (wait-for-input-available stream)
       (funcall function)))))

(defun thread-pids ()                   ;from sb-aclrepl
  "Return a list of the pids for all threads"
  (let ((offset (* 4 sb-vm::thread-pid-slot)))
    (sb-thread::mapcar-threads
     #'(lambda (sap) (sb-sys:sap-ref-32 sap offset)))))

(progn                                  ;from CLX dependent.lisp
  (declaim (inline yield))
  (defun yield ()
    (declare (optimize speed (safety 0)))
    (sb-alien:alien-funcall
     (sb-alien:extern-alien "sched_yield" (function sb-alien:int)))
    (values)))

(defun remove-fd-handler (handler)
  #+(or)
  (sb-sys:remove-fd-handler handler)
  #+(or) (sb-thread:terminate-thread handler) ;doesn't work
  (sb-thread:destroy-thread handler)
  (while (member handler (thread-pids))
    (yield)))
